__module__ = 'tlsverify.nist'
__version__ = 'NIST SP800-131A (strict mode)'

WEAK_KEY_SIZE = {
    'RSA': 2048,
    'DSA': 2048,
    'EC': 256,
    'DH': 256,
}
ALLOWED_CIPHERS = ['AES256-GCM-SHA384', 'AES256-SHA256', 'AES256-SHA', 'DES-CBC3-SHA', 'AES128-GCM-SHA256', 'AES128-SHA256', 'AES128-SHA']
WEAK_CIPHER_BITS = 160
NIST_NON_COMPLIANCE_WEAK_KEY_RSA = f'{__version__} compliance requires RSA public key bits to greater than {WEAK_KEY_SIZE["RSA"]}'
NIST_NON_COMPLIANCE_WEAK_KEY_DSA = f'{__version__} compliance requires DSA public key bits to greater than {WEAK_KEY_SIZE["DSA"]}'
NIST_NON_COMPLIANCE_WEAK_KEY_EC = f'{__version__} compliance requires Elliptic curve public key bits to greater than {WEAK_KEY_SIZE["EC"]}'
NIST_NON_COMPLIANCE_WEAK_KEY_DH = f'{__version__} compliance requires Diffie-Hellman exchange public key bits to greater than {WEAK_KEY_SIZE["DH"]}'
NIST_NON_COMPLIANCE_WEAK_PROTOCOL = f'{__version__} compliance requires deprecated and known weak TLS protocols are not supported'
NIST_NON_COMPLIANCE_CIPHER = f'{__version__} compliance requires cipher bits to greater than {WEAK_CIPHER_BITS}, not be of Anonymous key exchange suites, Exchange ciphers, or many block ciphers'
NIST_NON_COMPLIANCE_CA_TRUST = f'{__version__} compliance requires a complete Certificate chain with verified trust anchor'
NIST_NON_COMPLIANCE_MTLS = f'{__version__} compliance requires client certificate authentication'
VALIDATION_CA_TRUST = 'nist_ca_trust'
VALIDATION_WEAK_KEY = 'nist_weak_key'
VALIDATION_WEAK_CIPHER = 'nist_weak_cipher'
VALIDATION_WEAK_PROTOCOL = 'nist_weak_protocol'
VALIDATION_MTLS = 'nist_mtls'



__module__ = 'tlsverify.fips'
__version__ = 'FIPS 140-2 (NIST SP800-131A transition mode)'

WEAK_KEY_SIZE = {
    'RSA': 2048,
    'DSA': 2048,
    'EC': 224,
    'DH': 224,
}
ALLOWED_DEPRECATED_TLS_CIPHERS = ['AES256-SHA', 'DES-CBC3-SHA', 'AES128-SHA']
ALLOWED_CIPHERS = ['AES256-GCM-SHA384', 'AES256-SHA256', 'AES256-SHA', 'DES-CBC3-SHA', 'AES128-GCM-SHA256', 'AES128-SHA256', 'AES128-SHA']
WEAK_CIPHER_BITS = 160
FIPS_NON_COMPLIANCE_WEAK_KEY_RSA = f'{__version__} compliance requires RSA public key bits to greater than {WEAK_KEY_SIZE["RSA"]}'
FIPS_NON_COMPLIANCE_WEAK_KEY_DSA = f'{__version__} compliance requires DSA public key bits to greater than {WEAK_KEY_SIZE["DSA"]}'
FIPS_NON_COMPLIANCE_WEAK_KEY_EC = f'{__version__} compliance requires Elliptic curve public key bits to greater than {WEAK_KEY_SIZE["EC"]}'
FIPS_NON_COMPLIANCE_WEAK_KEY_DH = f'{__version__} compliance requires Diffie-Hellman exchange public key bits to greater than {WEAK_KEY_SIZE["DH"]}'
FIPS_NON_COMPLIANCE_WEAK_PROTOCOL = f'{__version__} compliance requires TLS protocol 1.2 support'
FIPS_NON_COMPLIANCE_CIPHER = f'{__version__} compliance requires cipher bits to greater than {WEAK_CIPHER_BITS}, not be of Anonymous key exchange suites, Exchange ciphers, or many block ciphers'
FIPS_NON_COMPLIANCE_CA_TRUST = f'{__version__} compliance requires a complete Certificate chain with verified trust anchor and every Certificate must meet the requirements (not just the leaf)'
FIPS_NON_COMPLIANCE_MTLS = f'{__version__} compliance requires TLS1.2, Client authentication violates the new FIPS restrictions and cannot be used for TLS1.0/1.1 in transition mode'
VALIDATION_CA_TRUST = 'fips_ca_trust'
VALIDATION_WEAK_KEY = 'fips_weak_key'
VALIDATION_WEAK_CIPHER = 'fips_weak_cipher'
VALIDATION_WEAK_PROTOCOL = 'fips_weak_protocol'
VALIDATION_MTLS = 'fips_mtls'



import json
import sys
import logging
import argparse
from pathlib import Path
from datetime import datetime
from dataclasses import asdict
import validators
from OpenSSL.crypto import X509
from rich import inspect
from rich.console import Console
from rich.style import Style
from rich.logging import RichHandler
from rich.progress import Progress
from rich.table import Table
from rich.style import Style
from rich import box
from . import exceptions, verify, util, validator, pci, nist, fips
from .transport import Transport


__version__ = 'tls-verify==0.4.14'
__module__ = 'tlsverify.cli'

CLI_COLOR_OK = 'dark_sea_green2'
CLI_COLOR_NOK = 'light_coral'
CLI_COLOR_ALERT = 'yellow3'
CLI_COLOR_NULL = 'magenta'
CLI_VALUE_TRUSTED = 'Trusted'
CLI_VALUE_NOT_TRUSTED = 'Not Trusted'
CLI_VALUE_VALID = 'Valid'
CLI_VALUE_NOT_VALID = 'Validation Errors'
CLI_VALUE_PASS = 'Pass'
CLI_VALUE_FAIL = 'Fail'
CLI_VALUE_YES = 'Yes'
CLI_VALUE_NO = 'No'
CLI_VALUE_DETECTED = 'Detected'
CLI_VALUE_PRESENT = 'Present'
CLI_VALUE_ABSENT = 'Absent'
CLI_VALUE_OK = 'OK'
CLI_VALUE_NOK = 'NOT OK'
CLI_VALUE_INEFFECTIVE = 'Ineffective'
CLI_VALUE_REVOKED = 'Revoked'
CLI_VALUE_NOT_REVOKED = 'Not Revoked'
CLI_VALUE_NA = 'N/A'
STYLES = {
    'certificate_valid': {'text': 'Certificate Valid', 'represent_as': (CLI_VALUE_VALID, CLI_VALUE_NOT_VALID), 'colors': (CLI_COLOR_OK, CLI_COLOR_NOK)},
    'certificate_chain_valid': {'text': 'Certificate Chain Valid', 'represent_as': (CLI_VALUE_VALID, CLI_VALUE_NOT_VALID), 'colors': (CLI_COLOR_OK, CLI_COLOR_NOK)},
    'certificate_chain_validation_result': {'text': 'Certificate Chain Validation Result'},
    validator.VALIDATION_CLIENT_AUTHENTICATION: {'text': 'Client Authentication', 'represent_as': (CLI_VALUE_PASS, CLI_VALUE_FAIL), 'colors': (CLI_COLOR_OK, CLI_COLOR_NOK)},
    validator.VALIDATION_CLIENT_AUTH_USAGE: {'text': 'Provided client Certificate authentication usage', 'represent_as': (CLI_VALUE_PASS, CLI_VALUE_FAIL), 'colors': (CLI_COLOR_OK, CLI_COLOR_NOK)},
    validator.VALIDATION_NOT_EXPIRED: {'text': 'Certificate is not expired', 'represent_as': (CLI_VALUE_PASS, CLI_VALUE_FAIL), 'colors': (CLI_COLOR_OK, CLI_COLOR_NOK)},
    validator.VALIDATION_ISSUED_PAST_TENSE: {'text': 'Certificate issued in the past', 'represent_as': (CLI_VALUE_PASS, CLI_VALUE_FAIL), 'colors': (CLI_COLOR_OK, CLI_COLOR_NOK)},
    validator.VALIDATION_SUBJECT_CN_DEFINED: {'text': 'Subject CN was defined', 'represent_as': (CLI_VALUE_PASS, CLI_VALUE_FAIL), 'colors': (CLI_COLOR_OK, CLI_COLOR_NOK)},
    validator.VALIDATION_SUBJECT_CN_VALID: {'text': 'Subject CN has valid syntax', 'represent_as': (CLI_VALUE_PASS, CLI_VALUE_FAIL), 'colors': (CLI_COLOR_OK, CLI_COLOR_NOK)},
    validator.VALIDATION_MATCH_HOSTNAME: {'text': 'Subject CN matches the server host name', 'represent_as': (CLI_VALUE_PASS, CLI_VALUE_FAIL), 'colors': (CLI_COLOR_OK, CLI_COLOR_NOK)},
    validator.VALIDATION_NOT_SELF_SIGNED: {'text': 'Not a self-signed Certificate', 'represent_as': (CLI_VALUE_PASS, CLI_VALUE_FAIL), 'colors': (CLI_COLOR_OK, CLI_COLOR_NOK)},
    validator.VALIDATION_WEAK_SIG_ALGO: {'text': 'Avoid known weak signature algorithms', 'represent_as': (CLI_VALUE_PASS, CLI_VALUE_FAIL), 'colors': (CLI_COLOR_OK, CLI_COLOR_NOK)},
    validator.VALIDATION_WEAK_KEYS: {'text': 'Avoid known weak public key algorithms', 'represent_as': (CLI_VALUE_PASS, CLI_VALUE_FAIL), 'colors': (CLI_COLOR_OK, CLI_COLOR_NOK)},
    validator.VALIDATION_DEPRECATED_TLS_PROTOCOLS: {'text': 'Avoid deprecated TLS protocols', 'represent_as': (CLI_VALUE_PASS, CLI_VALUE_FAIL), 'colors': (CLI_COLOR_OK, CLI_COLOR_NOK)},
    validator.VALIDATION_DEPRECATED_DNSSEC_ALGO: {'text': 'Avoid deprecated DNSSEC algorithms', 'represent_as': (CLI_VALUE_PASS, CLI_VALUE_FAIL), 'colors': (CLI_COLOR_OK, CLI_COLOR_NOK)},
    validator.VALIDATION_BASIC_CONSTRAINTS_CA: {'text': 'Leaf is not a CA', 'represent_as': (CLI_VALUE_PASS, CLI_VALUE_FAIL), 'colors': (CLI_COLOR_OK, CLI_COLOR_NOK)},
    validator.VALIDATION_VALID_TLS_USAGE: {'text': 'Key usage appropriate for TLS', 'represent_as': (CLI_VALUE_PASS, CLI_VALUE_FAIL), 'colors': (CLI_COLOR_OK, CLI_COLOR_NOK)},
    validator.VALIDATION_REVOCATION: {'text': 'Certificate chain not revoked', 'represent_as': (CLI_VALUE_PASS, CLI_VALUE_FAIL), 'colors': (CLI_COLOR_OK, CLI_COLOR_NOK)},
    validator.VALIDATION_ROOT_CA_TRUST: {'text': 'Root CA Certificate is trusted', 'represent_as': (CLI_VALUE_PASS, CLI_VALUE_FAIL), 'colors': (CLI_COLOR_OK, CLI_COLOR_NOK)},
    validator.VALIDATION_VALID_DNSSEC: {'text': 'DNSSEC Valid', 'represent_as': (CLI_VALUE_PASS, CLI_VALUE_FAIL), 'colors': (CLI_COLOR_OK, CLI_COLOR_NOK)},
    validator.VALIDATION_VALID_CAA: {'text': 'CAA Valid', 'represent_as': (CLI_VALUE_PASS, CLI_VALUE_FAIL), 'colors': (CLI_COLOR_OK, CLI_COLOR_NOK)},
    validator.VALIDATION_OCSP_STAPLE_SATISFIED: {'text': 'OCSP Staple satisfied', 'represent_as': (CLI_VALUE_PASS, CLI_VALUE_FAIL), 'colors': (CLI_COLOR_OK, CLI_COLOR_NOK)},
    validator.VALIDATION_OCSP_MUST_STAPLE_SATISFIED: {'text': 'OCSP Must Staple satisfied', 'represent_as': (CLI_VALUE_PASS, CLI_VALUE_FAIL), 'colors': (CLI_COLOR_OK, CLI_COLOR_NOK)},
    pci.VALIDATION_CA_TRUST: {'text': '[PCI] CA Trust', 'represent_as': (CLI_VALUE_PASS, CLI_VALUE_FAIL), 'colors': (CLI_COLOR_OK, CLI_COLOR_NOK)},
    pci.VALIDATION_WEAK_KEY: {'text': '[PCI] Key Size', 'represent_as': (CLI_VALUE_PASS, CLI_VALUE_FAIL), 'colors': (CLI_COLOR_OK, CLI_COLOR_NOK)},
    pci.VALIDATION_WEAK_CIPHER: {'text': '[PCI] Cipher bits', 'represent_as': (CLI_VALUE_PASS, CLI_VALUE_FAIL), 'colors': (CLI_COLOR_OK, CLI_COLOR_NOK)},
    pci.VALIDATION_WEAK_PROTOCOL: {'text': '[PCI] Deprecated protocols', 'represent_as': (CLI_VALUE_PASS, CLI_VALUE_FAIL), 'colors': (CLI_COLOR_OK, CLI_COLOR_NOK)},
    pci.VALIDATION_DEPRECATED_ALGO: {'text': '[PCI] Weak algorithms', 'represent_as': (CLI_VALUE_PASS, CLI_VALUE_FAIL), 'colors': (CLI_COLOR_OK, CLI_COLOR_NOK)},
    pci.VALIDATION_KNOWN_VULN_COMPRESSION: {'text': '[PCI] Vulnerable compression', 'represent_as': (CLI_VALUE_PASS, CLI_VALUE_FAIL), 'colors': (CLI_COLOR_OK, CLI_COLOR_NOK)},
    pci.VALIDATION_KNOWN_VULN_RENEGOTIATION: {'text': '[PCI] Vulnerable renegotiation', 'represent_as': (CLI_VALUE_PASS, CLI_VALUE_FAIL), 'colors': (CLI_COLOR_OK, CLI_COLOR_NOK)},
    pci.VALIDATION_KNOWN_VULN_SESSION_RESUMPTION: {'text': '[PCI] Vulnerable session resumption', 'represent_as': (CLI_VALUE_PASS, CLI_VALUE_FAIL), 'colors': (CLI_COLOR_OK, CLI_COLOR_NOK)},
    nist.VALIDATION_CA_TRUST: {'text': '[NIST] CA Trust', 'represent_as': (CLI_VALUE_PASS, CLI_VALUE_FAIL), 'colors': (CLI_COLOR_OK, CLI_COLOR_NOK)},
    nist.VALIDATION_WEAK_KEY: {'text': '[NIST] Key Size', 'represent_as': (CLI_VALUE_PASS, CLI_VALUE_FAIL), 'colors': (CLI_COLOR_OK, CLI_COLOR_NOK)},
    nist.VALIDATION_WEAK_CIPHER: {'text': '[NIST] Cipher bits', 'represent_as': (CLI_VALUE_PASS, CLI_VALUE_FAIL), 'colors': (CLI_COLOR_OK, CLI_COLOR_NOK)},
    nist.VALIDATION_WEAK_PROTOCOL: {'text': '[NIST] Deprecated protocols', 'represent_as': (CLI_VALUE_PASS, CLI_VALUE_FAIL), 'colors': (CLI_COLOR_OK, CLI_COLOR_NOK)},
    nist.VALIDATION_MTLS: {'text': '[NIST] Require ClientAuth', 'represent_as': (CLI_VALUE_PASS, CLI_VALUE_FAIL), 'colors': (CLI_COLOR_OK, CLI_COLOR_NOK)},
    fips.VALIDATION_CA_TRUST: {'text': '[FIPS] CA Trust', 'represent_as': (CLI_VALUE_PASS, CLI_VALUE_FAIL), 'colors': (CLI_COLOR_OK, CLI_COLOR_NOK)},
    fips.VALIDATION_WEAK_KEY: {'text': '[FIPS] Key Size', 'represent_as': (CLI_VALUE_PASS, CLI_VALUE_FAIL), 'colors': (CLI_COLOR_OK, CLI_COLOR_NOK)},
    fips.VALIDATION_WEAK_CIPHER: {'text': '[FIPS] Cipher bits', 'represent_as': (CLI_VALUE_PASS, CLI_VALUE_FAIL), 'colors': (CLI_COLOR_OK, CLI_COLOR_NOK)},
    fips.VALIDATION_WEAK_PROTOCOL: {'text': '[FIPS] Deprecated protocols', 'represent_as': (CLI_VALUE_PASS, CLI_VALUE_FAIL), 'colors': (CLI_COLOR_OK, CLI_COLOR_NOK)},
    fips.VALIDATION_MTLS: {'text': '[FIPS] No ClientAuth for TLS1.0/1.1', 'represent_as': (CLI_VALUE_PASS, CLI_VALUE_FAIL), 'colors': (CLI_COLOR_OK, CLI_COLOR_NOK)},
    'certificate_version': {'text': 'Certificate Version'},
    'certificate_public_key_type': {'text': 'Public Key Type'},
    'certificate_public_key_curve': {'text': 'Public Key Curve', 'null_as': CLI_VALUE_NA, 'null_color': CLI_COLOR_NULL},
    'certificate_public_key_size': {'text': 'Public Key Size'},
    'certificate_public_key_exponent': {'text': 'Public Key Exponent', 'null_as': CLI_VALUE_NA, 'null_color': CLI_COLOR_NULL},
    'certificate_private_key_pem': {'text': 'Derived private key (PEM format)'},
    'certificate_signature_algorithm': {'text': 'Signature Algorithm'},
    'certificate_pin_sha256': {'text': 'Certificate pin (sha256)'},
    'certificate_sha256_fingerprint': {'text': 'Fingerprint (sha256)'},
    'certificate_sha1_fingerprint': {'text': 'Fingerprint (sha1)'},
    'certificate_md5_fingerprint': {'text': 'Fingerprint (md5)'},
    'certificate_serial_number': {'text': 'Serial'},
    'certificate_serial_number_decimal': {'text': 'Serial (decimal)'},
    'certificate_serial_number_hex': {'text': 'Serial (hex)'},
    'certificate_subject': {'text': 'Certificate Subject'},
    'certificate_common_name': {'text': 'Certificate Subject CN'},
    'certificate_issuer': {'text': 'Issuer Subject CN'},
    'certificate_issuer_country': {'text': 'Issuer Country Code'},
    'certificate_not_before': {'text': 'Not Before'},
    'certificate_not_after': {'text': 'Not After'},
    'certificate_subject_key_identifier': {'text': 'Subject Key Identifier (SKI)'},
    'certificate_authority_key_identifier': {'text': 'Authority Key Identifier (AKI)'},
    'certificate_validation_type': {'text': 'Certificate Owner Validation Method'},
    'client_certificate_expected': {'text': 'Client Certificate Expected', 'represent_as': (CLI_VALUE_YES, CLI_VALUE_NO), 'colors': (CLI_COLOR_ALERT, CLI_COLOR_NULL)},
    'certification_authority_authorization': {'text': 'CAA', 'represent_as': (CLI_VALUE_PRESENT, CLI_VALUE_ABSENT), 'colors': (CLI_COLOR_OK, CLI_COLOR_NOK)},
    'revocation_ocsp_status': {'text': 'Revocation: OCSP'},
    'revocation_ocsp_response': {'text': 'Revocation: OCSP Response'},
    'revocation_ocsp_stapling': {'text': 'Revocation: OCSP Stapling', 'represent_as': (CLI_VALUE_PRESENT, CLI_VALUE_ABSENT), 'colors': (CLI_COLOR_OK, CLI_COLOR_NOK)},
    'revocation_ocsp_must_staple': {'text': 'Revocation: OCSP Must Staple flag', 'represent_as': (CLI_VALUE_PRESENT, CLI_VALUE_ABSENT), 'colors': (CLI_COLOR_OK, CLI_COLOR_NOK)},
    'revocation_ocsp_reason': {'text': 'Revocation: OCSP Revoked', 'null_as': CLI_VALUE_NOT_REVOKED, 'null_color': CLI_COLOR_OK, 'color': CLI_COLOR_NOK},
    'revocation_ocsp_time': {'text': 'Revocation: OCSP Revoked time', 'null_as': CLI_VALUE_NOT_REVOKED, 'null_color': CLI_COLOR_OK, 'color': CLI_COLOR_NOK},
    'revocation_crlite': {'text': 'Revocation: Mozilla CRLite', 'represent_as': (CLI_VALUE_REVOKED, CLI_VALUE_NOT_REVOKED), 'colors': (CLI_COLOR_NOK, CLI_COLOR_OK)},
    'sni_support': {'text': 'Server Name Indicator (SNI)', 'represent_as': (CLI_VALUE_PRESENT, CLI_VALUE_ABSENT), 'colors': (CLI_COLOR_OK, CLI_COLOR_NOK)},
    'negotiated_protocol': {'text': 'Negotiated Protocol'},
    'preferred_protocol': {'text': 'Server Preferred Protocol'},
    'offered_tls_versions': {'text': 'Offered TLS versions'},
    'negotiated_cipher': {'text': 'Negotiated Cipher'},
    'negotiated_cipher_bits': {'text': 'Negotiated Cipher bits'},
    'tls_version_intolerance': {'text': 'TLS version intolerance', 'represent_as': (CLI_VALUE_DETECTED, CLI_VALUE_OK), 'colors': (CLI_COLOR_NOK, CLI_COLOR_OK)},
    'tls_version_intolerance_versions': {'text': 'TLS version intolerance versions'},
    'tls_version_interference': {'text': 'TLS version interference', 'represent_as': (CLI_VALUE_DETECTED, CLI_VALUE_OK), 'colors': (CLI_COLOR_NOK, CLI_COLOR_OK)},
    'tls_version_interference_versions': {'text': 'TLS version interference versions'},
    'tls_long_handshake_intolerance': {'text': 'TLS long handshake intolerance', 'represent_as': (CLI_VALUE_DETECTED, CLI_VALUE_OK), 'colors': (CLI_COLOR_NOK, CLI_COLOR_OK)},
    'weak_cipher': {'text': 'Negotiated cipher is weak', 'represent_as': (CLI_VALUE_PRESENT, CLI_VALUE_OK), 'colors': (CLI_COLOR_NOK, CLI_COLOR_OK)},
    'strong_cipher': {'text': 'Strong negotiated cipher', 'represent_as': (CLI_VALUE_PRESENT, CLI_VALUE_NOK), 'colors': (CLI_COLOR_OK, CLI_COLOR_NOK)},
    'forward_anonymity': {'text': 'Forward Anonymity (FPS)', 'represent_as': (CLI_VALUE_PRESENT, CLI_VALUE_ABSENT), 'colors': (CLI_COLOR_OK, CLI_COLOR_NOK)},
    'session_resumption_caching': {'text': 'Session Resumption (caching)', 'represent_as': (CLI_VALUE_DETECTED, CLI_VALUE_OK), 'colors': (CLI_COLOR_NOK, CLI_COLOR_OK)},
    'session_resumption_tickets': {'text': 'Session Resumption (tickets)', 'represent_as': (CLI_VALUE_DETECTED, CLI_VALUE_OK), 'colors': (CLI_COLOR_NOK, CLI_COLOR_OK)},
    'session_resumption_ticket_hint': {'text': 'Session Resumption (ticket hint)', 'represent_as': (CLI_VALUE_DETECTED, CLI_VALUE_OK), 'colors': (CLI_COLOR_NOK, CLI_COLOR_OK)},
    'client_renegotiation': {'text': 'Insecure Client Renegotiation', 'represent_as': (CLI_VALUE_DETECTED, CLI_VALUE_OK), 'colors': (CLI_COLOR_NOK, CLI_COLOR_OK)},
    'compression_support': {'text': 'TLS Compression', 'represent_as': (CLI_VALUE_PRESENT, CLI_VALUE_OK), 'colors': (CLI_COLOR_NOK, CLI_COLOR_OK)},
    'dnssec': {'text': 'DNSSEC', 'represent_as': (CLI_VALUE_PRESENT, CLI_VALUE_ABSENT), 'colors': (CLI_COLOR_OK, CLI_COLOR_NOK)},
    'dnssec_algorithm': {'text': 'DNSSEC Algorithm', 'null_as': CLI_VALUE_NA, 'null_color': CLI_COLOR_NULL},
    'scsv': {'text': 'TLS downgrade prevention (SCSV)', 'represent_as': (CLI_VALUE_PRESENT, CLI_VALUE_INEFFECTIVE), 'colors': (CLI_COLOR_OK, CLI_COLOR_NOK)},
    'http_expect_ct_report_uri': {'text': 'Expect-CT report-uri', 'represent_as': (CLI_VALUE_OK, CLI_VALUE_ABSENT), 'colors': (CLI_COLOR_OK, CLI_COLOR_NOK)},
    'http_hsts': {'text': 'HSTS', 'represent_as': (CLI_VALUE_PRESENT, CLI_VALUE_ABSENT), 'colors': (CLI_COLOR_OK, CLI_COLOR_NOK)},
    'http_xfo': {'text': 'X-Frame-Options (XFO)', 'represent_as': (CLI_VALUE_PRESENT, CLI_VALUE_ABSENT), 'colors': (CLI_COLOR_OK, CLI_COLOR_NOK)},
    'http_csp': {'text': 'Content Security Policy (CSP)', 'represent_as': (CLI_VALUE_PRESENT, CLI_VALUE_ABSENT), 'colors': (CLI_COLOR_OK, CLI_COLOR_NOK)},
    'http_coep': {'text': 'Cross-Origin-Embedder-Policy (COEP)', 'represent_as': (CLI_VALUE_PRESENT, CLI_VALUE_ABSENT), 'colors': (CLI_COLOR_OK, CLI_COLOR_NOK)},
    'http_coop': {'text': 'Cross-Origin-Opener-Policy (COOP)', 'represent_as': (CLI_VALUE_PRESENT, CLI_VALUE_ABSENT), 'colors': (CLI_COLOR_OK, CLI_COLOR_NOK)},
    'http_corp': {'text': 'Cross-Origin-Resource-Policy (CORP)', 'represent_as': (CLI_VALUE_PRESENT, CLI_VALUE_ABSENT), 'colors': (CLI_COLOR_OK, CLI_COLOR_NOK)},
    'http_nosniff': {'text': 'X-Content-Type-Options nosniff', 'represent_as': (CLI_VALUE_PRESENT, CLI_VALUE_ABSENT), 'colors': (CLI_COLOR_OK, CLI_COLOR_NOK)},
    'http_unsafe_referrer': {'text': 'Referrer-Policy unsafe-url', 'represent_as': (CLI_VALUE_PRESENT, CLI_VALUE_OK), 'colors': (CLI_COLOR_NOK, CLI_COLOR_OK)},
    'http_xss_protection': {'text': 'X-XSS-Protection', 'represent_as': (CLI_VALUE_PRESENT, CLI_VALUE_ABSENT), 'colors': (CLI_COLOR_OK, CLI_COLOR_NOK)},
    'http_status_code': {'text': 'HTTP response status code'},
    'http1_support': {'text': 'HTTP/1', 'represent_as': (CLI_VALUE_YES, CLI_VALUE_NO), 'colors': (CLI_COLOR_NOK, CLI_COLOR_OK)},
    'http1_1_support': {'text': 'HTTP/1.1', 'represent_as': (CLI_VALUE_YES, CLI_VALUE_NO), 'colors': (CLI_COLOR_ALERT, CLI_COLOR_OK)},
    'http2_support': {'text': 'HTTP/2', 'represent_as': (CLI_VALUE_YES, CLI_VALUE_NO), 'colors': (CLI_COLOR_ALERT, CLI_COLOR_OK)},
    'http2_cleartext_support': {'text': 'HTTP/2 Cleartext', 'represent_as': (CLI_VALUE_YES, CLI_VALUE_NO), 'colors': (CLI_COLOR_NOK, CLI_COLOR_OK)},
    'possible_phish_or_malicious': {'text': 'Indicators of Phishing or Malware', 'represent_as': (CLI_VALUE_DETECTED, CLI_VALUE_OK), 'colors': (CLI_COLOR_NOK, CLI_COLOR_OK)},
    'trust_android': {'text': 'Android', 'represent_as': (CLI_VALUE_TRUSTED, CLI_VALUE_NOT_TRUSTED), 'colors': (CLI_COLOR_OK, CLI_COLOR_NOK)},
    'trust_android_status': {'text': ''},
    'trust_linux': {'text': 'Linux', 'represent_as': (CLI_VALUE_TRUSTED, CLI_VALUE_NOT_TRUSTED), 'colors': (CLI_COLOR_OK, CLI_COLOR_NOK)},
    'trust_linux_status': {'text': ''},
    'trust_java': {'text': 'Java', 'represent_as': (CLI_VALUE_TRUSTED, CLI_VALUE_NOT_TRUSTED), 'colors': (CLI_COLOR_OK, CLI_COLOR_NOK)},
    'trust_java_status': {'text': ''},
    'trust_certifi': {'text': 'Python', 'represent_as': (CLI_VALUE_TRUSTED, CLI_VALUE_NOT_TRUSTED), 'colors': (CLI_COLOR_OK, CLI_COLOR_NOK)},
    'trust_certifi_status': {'text': ''},
    'trust_ccadb': {'text': 'Common CA Database (CCADB)', 'represent_as': (CLI_VALUE_TRUSTED, CLI_VALUE_NOT_TRUSTED), 'colors': (CLI_COLOR_OK, CLI_COLOR_NOK)},
    'trust_ccadb_status': {'text': ''},
}
NEVER_SHOW = [
    'host',
    'port',
    "peer_address",
    "offered_ciphers",
    "certificate_san",
    "certificate_extensions",
    "certificate_is_self_signed",
    "certificate_private_key_pem",
    "subjectKeyIdentifier",
    "authorityKeyIdentifier",
    "certificate_expired",
]
JSON_ONLY = [
    "certificate_expired",
    "certificate_extensions",
    "certificate_is_self_signed",
]
SERVER_JSON_ONLY = JSON_ONLY + [
    'host',
    'port',
    "peer_address",
    "offered_ciphers",
]
SERVER_KEYS = [
    'certificate_validation_type',
    'certificate_is_self_signed',
    'certificate_subject',
    'certificate_issuer',
    'certificate_common_name',
    'certificate_intermediate_ca',
    'certificate_root_ca',
    'subjectKeyIdentifier',
    'authorityKeyIdentifier',
    'revocation_ocsp_status',
    'revocation_ocsp_response',
    'revocation_ocsp_reason',
    'revocation_ocsp_time',
    'revocation_ocsp_stapling',
    'revocation_ocsp_must_staple',
    'client_certificate_expected',
    'certification_authority_authorization',
    'dnssec',
    'dnssec_algorithm',
    'scsv',
    'compression_support',
    'tls_version_intolerance',
    'tls_version_intolerance_versions',
    'http_expect_ct_report_uri',
    'http_xss_protection',
    'http_status_code',
    'http1_support',
    'http1_1_support',
    'http2_support',
    'http2_cleartext_support',
    'sni_support',
    'negotiated_protocol',
    'preferred_protocol',
    'offered_tls_versions',
    'tls_version_interference',
    'tls_version_interference_versions',
    'tls_long_handshake_intolerance',
    'peer_address',
    'negotiated_cipher',
    'negotiated_cipher_bits',
    'weak_cipher',
    'strong_cipher',
    'forward_anonymity',
    'session_resumption_caching',
    'session_resumption_tickets',
    'session_resumption_ticket_hint',
    'client_renegotiation',
    'http_hsts',
    'http_xfo',
    'http_csp',
    'http_coep',
    'http_coop',
    'http_corp',
    'http_nosniff',
    'http_unsafe_referrer',
]
TRUST_KEYS = [
    'trust_ccadb',
    'trust_ccadb_status',
    'trust_java',
    'trust_java_status',
    'trust_android',
    'trust_android_status',
    'trust_linux',
    'trust_linux_status',
    'trust_certifi',
    'trust_certifi_status'
]
ROOT_SKIP = NEVER_SHOW + SERVER_KEYS + [
    'certificate_authority_key_identifier',
    'revocation_crlite',
    'common_name_defined',
    'possible_phish_or_malicious'
]
PEER_SKIP = NEVER_SHOW + SERVER_KEYS + TRUST_KEYS
SERVER_SKIP = NEVER_SHOW + TRUST_KEYS + [
    "certificate_root_ca",
    "certificate_intermediate_ca",
]
FINGERPRINTS = [
    'certificate_sha256_fingerprint',
    'certificate_sha1_fingerprint',
    'certificate_md5_fingerprint',
    'certificate_subject_key_identifier',
    'certificate_authority_key_identifier'
]
JSON_FILE = None

def _make_table(validator :validator.Validator, title :str, caption :str) -> Table:
    title_style = Style(bold=True, color=CLI_COLOR_OK if validator.certificate_valid else CLI_COLOR_NOK)
    table = Table(title=title, caption=caption, title_style=title_style, box=box.SIMPLE)
    table.add_column("", justify="right", style="dark_turquoise", no_wrap=True)
    table.add_column("Result", justify="left", no_wrap=False)
    return table

def _table_data(validator :validator.Validator, table :Table, skip :list[str]) -> Table:
    if 'verification_details' not in skip:
        for i, err in enumerate(validator.certificate_verify_messages):
            if any(key.startswith('pci_') for key in skip) and err.startswith('PCI'):
                continue
            if any(key.startswith('fips_') for key in skip) and err.startswith('FIPS'):
                continue
            if any(key.startswith('nist_') for key in skip) and err.startswith('NIST'):
                continue
            table.add_row(f'Note {i+1}', err)
    for key in validator.validation_checks.keys():
        if key in skip:
            continue
        table.add_row(STYLES.get(key,{}).get('text', key), util.styled_boolean(validator.validation_checks[key], STYLES[key]['represent_as'], STYLES[key]['colors']))
    for key in list(vars(validator.metadata).keys()):
        if key in skip:
            continue
        val = getattr(validator.metadata, key)
        if key in FINGERPRINTS and isinstance(val, str):
            table.add_row(STYLES.get(key,{}).get('text', key), util.str_n_split(val).upper())
            continue
        if val is None or (isinstance(val, str) and len(val) == 0):
            table.add_row(STYLES.get(key,{}).get('text', key), util.styled_value(STYLES[key].get('null_as', 'Unknown'), color=STYLES[key].get('null_color', CLI_COLOR_NULL)))
            continue
        if isinstance(val, str) and len(val) > 0:
            table.add_row(STYLES.get(key,{}).get('text', key), util.styled_value(val, color=STYLES[key].get('color')))
            continue
        if isinstance(val, bool):
            table.add_row(STYLES.get(key,{}).get('text', key), util.styled_boolean(val, represent_as=STYLES[key]['represent_as'], colors=STYLES[key]['colors']))
            continue
        table.add_row(STYLES.get(key,{}).get('text', key), util.styled_any(val))
    return table

def _table_ext(validator :validator.Validator, table :Table, skip :list[str]) -> Table:
    for v in validator.metadata.certificate_extensions:
        ext_data = v.copy()
        ext = ext_data['name']
        del ext_data['name']
        if ext in skip:
            continue
        if ext in ext_data:
            ext_sub = ext_data[ext]
            del ext_data[ext]
            table.add_row(f'Extension {ext}', util.styled_dict(ext_data))
            if isinstance(ext_sub, list):
                for sub in ext_sub:
                    if isinstance(sub, str):
                        table.add_row('', util.styled_value(sub))
                        continue
                    if isinstance(sub, dict):
                        for subk, subv in sub.items():
                            if subv is None:
                                table.add_row('', util.styled_value(subk))
                            else:
                                table.add_row('', subk+'='+util.styled_any(subv))
                        continue
                    table.add_row('', util.styled_any(sub))
                continue
            table.add_row('', str(ext_sub))
            continue    
        table.add_row(f'Extention {ext}', util.styled_any(ext_data))
    return table

def peer_outputs(validator :validator.PeerCertValidator) -> Table:
    peer_type = 'Intermediate Certificate'
    if validator.metadata.certificate_intermediate_ca:
        peer_type = 'Intermediate CA'
    title = f'{peer_type}: {validator.metadata.certificate_subject}'
    caption = '\n'.join([
        f'Issuer: {validator.metadata.certificate_issuer}',
        util.date_diff(validator.certificate.not_valid_after),
    ])
    table = _make_table(validator, title, caption)
    table.add_row(STYLES['certificate_valid']['text'], util.styled_boolean(validator.certificate_valid, STYLES['certificate_valid']['represent_as'], STYLES['certificate_valid']['colors']))
    _table_data(validator, table, PEER_SKIP)
    _table_ext(validator, table, PEER_SKIP)
    return table

def root_outputs(validator :validator.RootCertValidator) -> Table:
    title = f'Root CA: {validator.metadata.certificate_subject}'
    caption = util.date_diff(validator.certificate.not_valid_after)
    table = _make_table(validator, title, caption)
    table.add_row(STYLES['certificate_valid']['text'], util.styled_boolean(validator.certificate_valid, ('Trusted', 'Not Trusted'), STYLES['certificate_valid']['colors']))
    _table_data(validator, table, ROOT_SKIP)
    _table_ext(validator, table, ROOT_SKIP)
    return table

def server_outputs(validator :validator.CertValidator) -> Table:
    title = f'Leaf Certificate {validator.metadata.host}:{validator.metadata.port} ({validator.metadata.peer_address})'
    caption = '\n'.join([
        f'Issuer: {validator.metadata.certificate_issuer}',
        util.date_diff(validator.certificate.not_valid_after),
    ])
    table = _make_table(validator, title, caption)
    table.add_row(STYLES['certificate_valid']['text'], util.styled_boolean(validator.certificate_valid, STYLES['certificate_valid']['represent_as'], STYLES['certificate_valid']['colors']))
    table.add_row(STYLES['certificate_chain_valid']['text'], util.styled_boolean(validator.certificate_chain_valid, STYLES['certificate_chain_valid']['represent_as'], STYLES['certificate_chain_valid']['colors']))
    table.add_row(STYLES['certificate_chain_validation_result']['text'], util.styled_any(validator.certificate_chain_validation_result))
    _table_data(validator, table, SERVER_SKIP)
    _table_ext(validator, table, SERVER_SKIP)
    return table

def update_bar(progress, task):
    def progress_bar(completed :int = None):
        if isinstance(completed, int):
            progress.update(task, completed=completed)
        else:
            progress.update(task, advance=1)
    return progress_bar

def output(result, debug :bool = False):
    console = Console()
    if debug and hasattr(result, 'transport'):
        inspect(result.transport, title=result.transport.negotiated_protocol)
    if debug and hasattr(result, 'metadata'):
        inspect(result.metadata, title=result.metadata.certificate_subject)
    if isinstance(result, validator.RootCertValidator):
        console.print(root_outputs(result))
    if isinstance(result, validator.PeerCertValidator):
        console.print(peer_outputs(result))
    if isinstance(result, validator.CertValidator):
        console.print(server_outputs(result))
    console.print('\n\n')

def validator_data(result :validator.Validator, certificate_type :str, skip_keys :list) -> dict:
    data = asdict(result.metadata)
    data['certificate_valid'] = result.certificate_valid
    if isinstance(result, validator.CertValidator):
        data['certificate_chain_valid'] = result.certificate_chain_valid
        data['certificate_chain_validation_result'] = result.certificate_chain_validation_result
    data['certificate_type'] = certificate_type
    data['expiry_status'] = util.date_diff(result.certificate.not_valid_after)
    data['verification_results'] = {}
    for key, value in result.validation_checks.items():
        if key in skip_keys:
            continue
        data['verification_results'][key] = value
    if 'verification_details' not in skip_keys:
        data['verification_details'] = result.certificate_verify_messages
    if any(key.startswith('pci_') for key in skip_keys):
        data['verification_details'] = [detail for detail in data['verification_details'] if not detail.startswith('PCI')]
    if any(key.startswith('nist_') for key in skip_keys):
        data['verification_details'] = [detail for detail in data['verification_details'] if not detail.startswith('NIST')]
    if any(key.startswith('fips_') for key in skip_keys):
        data['verification_details'] = [detail for detail in data['verification_details'] if not detail.startswith('FIPS')]
    for key in list(vars(result.metadata).keys()):
        if key in skip_keys and key in data:
            del data[key]
    for v in result.metadata.certificate_extensions:
        if v.get('name') in skip_keys:
            data['certificate_extensions'][:] = [d for d in data['certificate_extensions'] if d.get('name') != v['name']]

    return data

def make_json(results :list[validator.Validator], evaluation_duration_seconds :int) -> str:
    data = {
        'generator': __version__,
        'date': datetime.utcnow().replace(microsecond=0).isoformat(),
        'evaluation_duration_seconds': evaluation_duration_seconds,
        'validations': []
    }
    for result in results:
        if isinstance(result, validator.RootCertValidator):
            data['validations'].append(validator_data(result, 'Root CA', [x for x in ROOT_SKIP if x not in JSON_ONLY]))
        if isinstance(result, validator.PeerCertValidator):
            cert_type = 'Intermediate Certificate'
            if result.metadata.certificate_intermediate_ca:
                cert_type = 'Intermediate CA'
            data['validations'].append(validator_data(result, cert_type, [x for x in PEER_SKIP if x not in JSON_ONLY]))
        if isinstance(result, validator.CertValidator):
            data['validations'].append(validator_data(result, 'Leaf Certificate', [x for x in SERVER_SKIP if x not in SERVER_JSON_ONLY]))
    return json.dumps(data, sort_keys=True, default=str)

def with_progress_bars(domains :list[tuple[str, int]], cafiles :list = None, use_sni :bool = True, client_pem :str = None, tmp_path_prefix :str = '/tmp', debug :bool = False) -> list[validator.Validator]:
    if not isinstance(client_pem, str) and client_pem is not None:
        raise TypeError(f"provided an invalid type {type(client_pem)} for client_pem, expected list")
    if not isinstance(cafiles, list) and cafiles is not None:
        raise TypeError(f"provided an invalid type {type(cafiles)} for cafiles, expected list")
    if not isinstance(use_sni, bool):
        raise TypeError(f"provided an invalid type {type(use_sni)} for tlsext, expected list")
    if not isinstance(tmp_path_prefix, str):
        raise TypeError(f"provided an invalid type {type(tmp_path_prefix)} for tmp_path_prefix, expected str")

    evaluation_start = datetime.utcnow()
    results = []
    with Progress() as progress:
        prog_client_auth = progress.add_task("[cyan]Client Authentication...", total=5*len(domains))
        prog_tls = progress.add_task("[cyan]Evaluating TLS...", total=14*len(domains))
        prog_cert_val = progress.add_task("[cyan]Certificate Chain Validation...", total=7*len(domains))
        while not progress.finished:
            for host, port in domains:
                if not isinstance(port, int):
                    raise TypeError(f"provided an invalid type {type(port)} for port, expected int")
                if validators.domain(host) is not True:
                    raise ValueError(f"provided an invalid domain {host}")
                result = validator.CertValidator()
                transport = Transport(host, port)
                if client_pem is None:
                    progress.update(prog_client_auth, visible=False)
                else:
                    progress.update(prog_client_auth, advance=1)
                    transport.pre_client_authentication_check(client_pem_path=client_pem, progress_bar=update_bar(progress, prog_client_auth))
                    progress.update(prog_client_auth, advance=1)
                if not transport.connect_least_secure(cafiles=cafiles, use_sni=use_sni, progress_bar=update_bar(progress, prog_tls)) or not isinstance(transport.server_certificate, X509):
                    raise exceptions.ValidationError(exceptions.VALIDATION_ERROR_TLS_FAILED.format(host=host, port=port))
                progress.update(prog_tls, advance=1)
                if isinstance(tmp_path_prefix, str):
                    result.tmp_path_prefix = tmp_path_prefix
                result.mount(transport)
                progress.update(prog_cert_val, advance=1)
                result.verify()
                progress.update(prog_cert_val, advance=1)
                result.verify_chain(progress_bar=update_bar(progress, prog_cert_val))
                result.pcidss_compliant()
                result.fips_compliant()
                result.nist_compliant()
                results += result.peer_validations
                results.append(result)
            progress.update(prog_client_auth, completed=5*len(domains))
            progress.update(prog_tls, completed=14*len(domains))
            progress.update(prog_cert_val, completed=7*len(domains))

    console = Console()
    valid = all(v.certificate_valid for v in results)
    for result in results:
        output(result, debug=debug)
    result_style = Style(color='dark_sea_green2' if valid else 'light_coral')
    console.print('Valid ✓✓✓' if valid else '\nNot Valid. There where validation errors', style=result_style)
    console.print(f'Evaluation duration seconds {(datetime.utcnow() - evaluation_start).total_seconds()}\n\n')
    return results

def cli():
    parser = argparse.ArgumentParser()
    parser.add_argument("targets", nargs="*", help='All unnamed arguments are hosts (and ports) targets to test. ~$ tlsverify google.com:443 github.io owasp.org:80')
    parser.add_argument('-H', '--host', help='single host to check', dest='host', default=None)
    parser.add_argument('-p', '--port', help='TLS port of host', dest='port', default=443)
    parser.add_argument('-c', '--cafiles', help='path to PEM encoded CA bundle file, url or file path accepted', dest='cafiles', default=None)
    parser.add_argument('-C', '--client-pem', help='path to PEM encoded client certificate, url or file path accepted', dest='client_pem', default=None)
    parser.add_argument('-t', '--tmp-path-prefix', help='local file path to use as a prefix when saving temporary files such as those being fetched for client authorization', dest='tmp_path_prefix', default='/tmp')
    parser.add_argument('--pci-dss', help='Include PCI DSS requirements assertions', dest='show_pci', action="store_true")
    parser.add_argument('--nist-strict-mode', help='Include NIST SP800-131A strict mode assertions', dest='show_nist', action="store_true")
    parser.add_argument('--fips-nist-transition-mode', help='Include FIPS 140-2 transition to NIST SP800-131A assertions', dest='show_fips', action="store_true")
    parser.add_argument('--disable-sni', help='Do not negotiate SNI using INDA encoded host', dest='disable_sni', action="store_true")
    parser.add_argument('--show-private-key', help='If the private key is exposed, show it in the results', dest='show_private_key', action="store_true")
    parser.add_argument('--hide-validation-details', help='Do not include detailed validation messages in output', dest='hide_validation_details', action="store_true")
    parser.add_argument('-j', '--json-file', help='Store to file as JSON', dest='json_file', default=None)
    parser.add_argument('--hide-progress-bars', help='Hide task progress bars', dest='hide_progress_bars', action="store_true")
    parser.add_argument('-v', '--errors-only', help='set logging level to ERROR (default CRITICAL)', dest='log_level_error', action="store_true")
    parser.add_argument('-vv', '--warning', help='set logging level to WARNING (default CRITICAL)', dest='log_level_warning', action="store_true")
    parser.add_argument('-vvv', '--info', help='set logging level to INFO (default CRITICAL)', dest='log_level_info', action="store_true")
    parser.add_argument('-vvvv', '--debug', help='set logging level to DEBUG (default CRITICAL)', dest='log_level_debug', action="store_true")
    parser.add_argument('--version', dest='show_version', action="store_true")
    args = parser.parse_args()
    JSON_FILE = args.json_file
    log_level = logging.CRITICAL
    if args.log_level_error:
        log_level = logging.ERROR
    if args.log_level_warning:
        log_level = logging.WARNING
    if args.log_level_info:
        log_level = logging.INFO
    if args.log_level_debug:
        log_level = logging.DEBUG
    handlers = []
    log_format = '%(asctime)s - %(name)s - [%(levelname)s] %(message)s'
    if sys.stdout.isatty():
        log_format = '%(message)s'
        handlers.append(RichHandler(rich_tracebacks=True))
    logging.basicConfig(
        format=log_format,
        level=log_level,
        handlers=handlers
    )
    debug = log_level==logging.DEBUG
    def version(): import platform; print(f"{__version__} Python {sys.version} {platform.platform()} {platform.uname().node} {platform.uname().release} {platform.version()}")
    if args.show_version:
        version()
        sys.exit(0)
    if args.host is None and len(args.targets) == 0:
        version()
        parser.print_help(sys.stderr)
        sys.exit(1)
    if args.host is not None:
        args.targets.append(f'{args.host}:{args.port}')

    domains = []
    for target in args.targets:
        pieces = target.split(':')
        host, port = None, None
        if len(pieces) == 2:
            host, port = pieces
        if len(pieces) == 1:
            host = pieces[0]
            port = args.port
        if validators.domain(host) is not True:
            raise AttributeError(f'host {host} is invalid')
        domains.append((host, int(port)))

    if args.hide_validation_details:
        SERVER_SKIP.append('verification_details')
        PEER_SKIP.append('verification_details')
        ROOT_SKIP.append('verification_details')
    if args.show_private_key:
        SERVER_SKIP.remove('certificate_private_key_pem')
        PEER_SKIP.remove('certificate_private_key_pem')
        ROOT_SKIP.remove('certificate_private_key_pem')

    if not args.show_nist:
        nist_keys = [
            nist.VALIDATION_CA_TRUST,
            nist.VALIDATION_WEAK_KEY,
            nist.VALIDATION_WEAK_CIPHER,
            nist.VALIDATION_WEAK_PROTOCOL,
            nist.VALIDATION_MTLS,
        ]
        SERVER_SKIP.extend(nist_keys)
        PEER_SKIP.extend(nist_keys)
        ROOT_SKIP.extend(nist_keys)

    if not args.show_fips:
        fips_keys = [
            fips.VALIDATION_CA_TRUST,
            fips.VALIDATION_WEAK_KEY,
            fips.VALIDATION_WEAK_CIPHER,
            fips.VALIDATION_WEAK_PROTOCOL,
            fips.VALIDATION_MTLS,
        ]
        SERVER_SKIP.extend(fips_keys)
        PEER_SKIP.extend(fips_keys)
        ROOT_SKIP.extend(fips_keys)

    if not args.show_pci:
        pci_keys = [
            pci.VALIDATION_CA_TRUST,
            pci.VALIDATION_WEAK_KEY,
            pci.VALIDATION_WEAK_CIPHER,
            pci.VALIDATION_WEAK_PROTOCOL,
            pci.VALIDATION_DEPRECATED_ALGO,
            pci.VALIDATION_KNOWN_VULN_COMPRESSION,
            pci.VALIDATION_KNOWN_VULN_RENEGOTIATION,
            pci.VALIDATION_KNOWN_VULN_SESSION_RESUMPTION
        ]
        SERVER_SKIP.extend(pci_keys)
        PEER_SKIP.extend(pci_keys)
        ROOT_SKIP.extend(pci_keys)

    all_results = []
    evaluation_start = datetime.utcnow()
    if args.hide_progress_bars:
        for domain, port in domains:
            _, results = verify(
                domain,
                int(port),
                cafiles=args.cafiles,
                use_sni=not args.disable_sni,
                client_pem=args.client_pem,
                tmp_path_prefix=args.tmp_path_prefix,
            )
            all_results += results
        console = Console()
        valid = all([v.certificate_valid for v in all_results])
        for result in all_results:
            output(result, debug=debug)
        result_style = Style(color=CLI_COLOR_OK if valid else CLI_COLOR_NOK)
        console.print('Valid ✓✓✓' if valid else '\nNot Valid. There where validation errors', style=result_style)
        console.print(f'Evaluation duration seconds {(datetime.utcnow() - evaluation_start).total_seconds()}\n\n')
    else:
        all_results = with_progress_bars( # clones tlsverify.verify and only adds progress bars
            domains=domains,
            cafiles=args.cafiles,
            use_sni=not args.disable_sni,
            client_pem=args.client_pem,
            tmp_path_prefix=args.tmp_path_prefix,
            debug=debug
        )

    if JSON_FILE:
        json_path = Path(JSON_FILE)
        if json_path.is_file():
            json_path.unlink()
        json_path.write_text(make_json(all_results, (datetime.utcnow() - evaluation_start).total_seconds()))



import logging
import ssl
from base64 import urlsafe_b64encode
from datetime import datetime
from time import sleep
from socket import socket, AF_INET, SOCK_STREAM, MSG_PEEK
from pathlib import Path
from cryptography.x509 import extensions, oid
from cryptography.x509.base import Certificate
from cryptography.x509.ocsp import OCSPResponse, load_der_ocsp_response, OCSPResponseStatus, OCSPCertStatus, OCSPRequestBuilder
from cryptography.hazmat.primitives.hashes import SHA1
from cryptography.hazmat.primitives.serialization import Encoding
from OpenSSL import SSL, _util
from OpenSSL.SSL import _lib as native_openssl
from OpenSSL.crypto import X509, FILETYPE_PEM, X509Name, load_certificate
from certifi import where
from hyperframe.frame import Frame, SettingsFrame
from hyperframe.exceptions import InvalidFrameError
from requests import post
import validators
import idna
from . import util
from . import exceptions


__module__ = 'tlsverify.metadata'
logger = logging.getLogger(__name__)

FAKE_PROTOCOLS = ['TLSv1.4', 'TLSv1.8', 'TLSv2', 'TLSv2.1', 'TLSv2.3']

class Transport:
    host :str
    port :int
    client_pem_path :str
    negotiated_protocol :str
    negotiated_cipher :str
    negotiated_cipher_bits :int
    peer_address :str
    sni_support :bool
    http2_support :bool
    http2_cleartext_support :bool
    http2_response_frame :str
    http1_support :bool
    http1_code :int
    http1_status :str
    http1_response_proto :str
    http1_headers :dict
    http1_1_support :bool
    http1_1_code :int
    http1_1_status :str
    http1_1_response_proto :str
    http1_1_headers :dict
    ocsp_stapling : bool
    ocsp_must_staple : bool
    ocsp_revocation_reason :str
    ocsp_revocation_time :str
    ocsp_response_status :str
    ocsp_certificate_status :str
    ciphers :list
    compression :str
    client_renegotiation :bool
    session_cache_mode :str
    session_tickets :bool
    session_ticket_hints :bool
    server_certificate :X509
    client_certificate :X509
    client_certificate_expected :bool
    client_certificate_match :bool
    tls_downgrade :bool
    offered_tls_versions :list
    tls_version_intolerance :bool
    tls_version_intolerance_versions :list
    tls_version_interference :bool
    tls_version_interference_versions :list
    long_handshake_intolerance :bool
    preferred_protocol :str
    cafiles :list
    certificate_chain :list[X509]
    verifier_errors :list[tuple[X509], int]
    expected_client_subjects :list[X509Name]
    tmp_path_prefix :str = '/tmp'
    default_connect_method :str = 'SSLv23_METHOD'
    default_connect_verify_mode :str = 'VERIFY_NONE'
    data_recv_size :int = 8096
    depth :dict

    def __init__(self, host :str, port :int = 443) -> None:
        if not isinstance(port, int):
            raise TypeError(f"provided an invalid type {type(port)} for port, expected int")
        self.port = port
        if validators.domain(host) is not True:
            raise ValueError(f"provided an invalid domain {host}")
        self.host = host
        self.client_pem_path = None
        self.client_pem_path = None
        self.negotiated_protocol = None
        self.negotiated_cipher = None
        self.negotiated_cipher_bits = None
        self.ciphers = []
        self.client_renegotiation = None
        self.compression = None
        self.session_cache_mode = None
        self.session_tickets = None
        self.session_ticket_hints = None
        self.server_certificate = None
        self.client_certificate = None
        self.client_certificate_expected = None
        self.client_certificate_match = None
        self.certificate_chain = []
        self.verifier_errors = []
        self.cafiles = []
        self.expected_client_subjects = []
        self.sni_support = False
        self.http1_support = False
        self.http1_1_support = False
        self.http2_support = False
        self.http2_cleartext_support = False
        self.peer_address = None
        self.tls_downgrade = None
        self.offered_tls_versions = []
        self.tls_version_interference = None
        self.tls_version_interference_versions = []
        self.tls_version_intolerance = None
        self.tls_version_intolerance_versions = []
        self.long_handshake_intolerance = None
        self.preferred_protocol = None
        self.ocsp_stapling = None
        self.ocsp_must_staple = None
        self.ocsp_revocation_reason = None
        self.ocsp_revocation_time = None
        self.ocsp_response_status = None
        self.ocsp_certificate_status = None
        self.depth = {}

    def pre_client_authentication_check(self, client_pem_path :str = None, progress_bar :callable = lambda *args: None) -> bool:
        if not isinstance(self.port, int):
            raise TypeError(f"provided an invalid type {type(self.port)} for port, expected int")
        if validators.domain(self.host) is not True:
            raise ValueError(f"provided an invalid domain {self.host}")
        if not isinstance(client_pem_path, str):
            raise TypeError(f"provided an invalid type {type(client_pem_path)} for client_pem_path, expected str")

        self.client_certificate_expected = True
        self.client_certificate_match = False
        valid_client_pem = util.filter_valid_files_urls([client_pem_path], self.tmp_path_prefix)
        if valid_client_pem is False:
            logger.error('client_pem_path was provided but is not a valid URL or file does not exist')
            return False
        if isinstance(valid_client_pem, list) and len(valid_client_pem) == 1:
            self.client_pem_path = valid_client_pem[0]
        logger.info('Negotiating with the server to derive expected client certificate subjects')
        ctx = SSL.Context(method=getattr(SSL, self.default_connect_method))
        ctx.load_verify_locations(cafile=where())
        ctx.verify_mode = SSL.VERIFY_NONE
        ctx.check_hostname = False
        conn = SSL.Connection(ctx, self.prepare_socket())
        conn.connect((self.host, self.port))
        if ssl.HAS_SNI:
            conn.set_tlsext_host_name(idna.encode(self.host))
        conn.setblocking(1)
        conn.set_connect_state()
        progress_bar()
        util.do_handshake(conn)
        self.expected_client_subjects = conn.get_client_ca_list()
        conn.close()
        progress_bar()
        if len(self.expected_client_subjects) > 0:
            logger.info('Checking client certificate')
            self.client_certificate = load_certificate(FILETYPE_PEM, Path(self.client_pem_path).read_bytes())
            logger.debug(f'issuer subject: {self.client_certificate.get_issuer().commonName}')
            for check in self.expected_client_subjects:
                logger.debug(f'expected subject: {check.commonName}')
                if self.client_certificate.get_issuer().commonName == check.commonName:
                    self.client_certificate_match = True
                    progress_bar()
                    return True
        return False

    def _verifier(self, conn :SSL.Connection, server_cert :X509, errno :int, depth :int, preverify_ok :int):
        # preverify_ok indicates, whether the verification of the server certificate in question was passed (preverify_ok=1) or not (preverify_ok=0)
        # https://www.openssl.org/docs/man1.0.2/man1/verify.html
        if errno in exceptions.X509_MESSAGES.keys():
            self.verifier_errors.append((server_cert, errno))
        return True

    def do_request(self, conn :SSL.Connection, method :str = 'HEAD', uri_path :str = '/', protocol :str = 'HTTP/1.1', request_compression :bool = True):
        if method.upper() not in ['HEAD', 'GET', 'OPTIONS']:
            raise AttributeError(f'method {method} not supported')
        if protocol.upper() not in ['HTTP/1.0', 'HTTP/1.1']:
            raise AttributeError(f'protocol {protocol} not supported')
        if not isinstance(uri_path, str):
            raise AttributeError(f'uri_path not supported')
        if not uri_path.startswith('/'):
            uri_path = f'/{uri_path}'

        request = [
            f"{method.upper()} {uri_path} {protocol}",
            f"Host: {self.host}",
            "Accept: */*",
            "Cache-Control: max-age=0",
            "Connection: close",
            "Content-Length: 0",
            "User-Agent: pypi.org/project/tls-verify/",
        ]
        if request_compression is True:
            request.append("Accept-Encoding: compress, gzip")
        request = "\r\n".join(request) + "\r\n\r\n"
        logger.info(f'Request:\n{request}')
        conn.sendall(request.encode())
        head = b""
        try:
            while not head.endswith(b"\r\n\r\n"):
                if Transport.is_connection_closed(conn): break
                head += conn.recv(1)
        except ConnectionResetError:
            pass
        except SSL.ZeroReturnError:
            pass
        if method.upper() in ['HEAD', 'OPTIONS'] or protocol.upper() == 'HTTP/2':
            return head.decode(), None
        body = b""
        try:
            while b"\r\n\r\n" not in body:
                if Transport.is_connection_closed(conn): break
                data = conn.recv(Transport.data_recv_size)
                if len(data) == 0: break
                body += data
        except ConnectionResetError:
            pass
        except SSL.ZeroReturnError:
            pass
        return head.decode(), body.decode()

    def _protocol_handler(self, conn :SSL.Connection, protocol :str ='HTTP/1.1'):
        proto_map = {
            'HTTP/1.0': 'http1_',
            'HTTP/1.1': 'http1_1_',
        }
        logger.debug(f'protocol {protocol}')
        head, _ = self.do_request(conn, protocol=protocol)
        logger.info(f'Response headers:\n{head}')
        header = Transport.parse_header(head)
        prefix = proto_map[protocol]
        response_code = int(header['response_code'])
        setattr(self, f'{prefix}support', header['protocol'].startswith(protocol))
        setattr(self, f'{prefix}code', response_code)
        setattr(self, f'{prefix}status', header['response_status'])
        setattr(self, f'{prefix}response_proto', header['protocol'])
        setattr(self, f'{prefix}headers', header['headers'])
        if self.client_renegotiation is None:
            total_renegotiations = conn.total_renegotiations()
            proceed = conn.renegotiate()
            if proceed:
                try:
                    conn.setblocking(0)
                    util.do_handshake(conn)
                    self.client_renegotiation = conn.total_renegotiations() > total_renegotiations
                except SSL.ZeroReturnError: pass
                except Exception as ex:
                    logger.warning(ex, exc_info=True)
            self.client_renegotiation = self.client_renegotiation is True
        return True

    def _get_ocsp_response(self, issuer :Certificate, uri :str, timeout :int = 3) -> OCSPResponse:
        builder = OCSPRequestBuilder()
        builder = builder.add_certificate(self.server_certificate.to_cryptography(), issuer, SHA1())
        ocsp_request = builder.build()
        try:
            response = post(
                uri,
                data=ocsp_request.public_bytes(Encoding.DER),
                headers={'Content-Type': 'application/ocsp-request'},
                timeout=timeout)
        except Exception as ex:
            logger.warning(ex, exc_info=True)
            return None
        if response.status_code != 200:
            logger.warning("HTTP request returned %d", response.status_code)
            return None
        response = load_der_ocsp_response(response.content)
        if response.response_status != OCSPResponseStatus.SUCCESSFUL:
            return None
        if response.serial_number != ocsp_request.serial_number:
            logger.debug("Response serial number does not match request")
            return None
        return response

    def _ocsp_handler(self, conn :SSL.Connection, assertion :bytes, userdata) -> bool:
        if not isinstance(self.server_certificate, X509):
            self.server_certificate = conn.get_peer_certificate()
            for (_, cert) in enumerate(conn.get_peer_cert_chain()):
                self.certificate_chain.append(cert)
        issuer = util.issuer_from_chain(self.server_certificate, self.certificate_chain)
        if not isinstance(issuer, X509):
            logger.warning('Issuer certificate not found in chain')
            return False
        self.ocsp_stapling = False
        self.ocsp_must_staple = False
        ext = None
        try:
            ext = self.server_certificate.to_cryptography().extensions.get_extension_for_class(extensions.TLSFeature)
        except Exception:
            pass
        if ext is not None:
            for feature in ext.value:
                if feature == extensions.TLSFeatureType.status_request:
                    logger.debug("Peer presented a must-staple cert")
                    self.ocsp_must_staple = True
                    break
        response = None
        if assertion == b'':
            if self.ocsp_must_staple is True:
                return False # stapled response is expected and required
            ext = self.server_certificate.to_cryptography().extensions.get_extension_for_class(extensions.AuthorityInformationAccess)
            if ext is None:
                return True # stapled response is expected though not required, not very good but still a valid assertion
            uris = [desc.access_location.value for desc in ext.value if desc.access_method == oid.AuthorityInformationAccessOID.OCSP]
            if not uris:
                return True # stapled response is expected though not required, without any responders it is still a valid assertion
            for uri in uris:
                logger.debug(f"Requesting OCSP from responder {uri}")
                response = self._get_ocsp_response(issuer.to_cryptography(), uri)
                if response is None:
                    continue
        if response is None and assertion != b'':
            self.ocsp_stapling = True
            response = load_der_ocsp_response(assertion)
        if response is None:
            logger.warning("OCSP response is not available")
            return False
        if response.this_update > datetime.utcnow():
            logger.error("OCSP thisUpdate is future dated")
            return False
        logger.info("OCSP response received")
        if response.revocation_reason:
            self.ocsp_revocation_reason = response.revocation_reason.value
        if response.revocation_time:
            self.ocsp_revocation_time = response.revocation_time.value
        if response.response_status.value in util.OCSP_RESP_STATUS:
            self.ocsp_response_status = util.OCSP_RESP_STATUS[response.response_status.value]
        if response.certificate_status.value in util.OCSP_CERT_STATUS:
            self.ocsp_certificate_status = util.OCSP_CERT_STATUS[response.certificate_status.value]
        return response.response_status == OCSPResponseStatus.SUCCESSFUL and response.certificate_status == OCSPCertStatus.GOOD

    def prepare_socket(self, timeout :int = 1):
        sock = socket(AF_INET, SOCK_STREAM)
        sock.settimeout(timeout)
        return sock

    def prepare_context(self, method :str = None, verify_mode :str = None, check_hostname :bool = False) -> SSL.Context:
        if not isinstance(check_hostname, bool):
            raise TypeError(f'check_hostname {type(check_hostname)}, bool supported')
        if method is not None and not isinstance(method, str):
            raise TypeError(f'method {type(method)}, str supported')
        if method is None:
            method = Transport.default_connect_method
        if isinstance(method, str) and not hasattr(SSL, method):
            raise AttributeError('Only available SSL methods on your system are supported')
        if verify_mode is not None and not isinstance(verify_mode, str):
            raise TypeError(f'verify_mode {type(verify_mode)}, str supported')
        if verify_mode is None:
            verify_mode = Transport.default_connect_verify_mode
        if isinstance(verify_mode, str) and not hasattr(SSL, verify_mode):
            raise AttributeError('Only available SSL verify modes on your system are supported')
        ctx = SSL.Context(method=getattr(SSL, method))
        ctx.load_verify_locations(cafile=where())
        for cafile in self.cafiles:
            ctx.load_verify_locations(cafile=cafile)
        if self.client_certificate_expected and isinstance(self.client_pem_path, str):
            ctx.use_certificate_file(certfile=self.client_pem_path, filetype=FILETYPE_PEM)
        ctx.verify_mode = verify_mode
        ctx.check_hostname = check_hostname
        return ctx

    def prepare_connection(self, context :SSL.Context, sock :socket = None, use_sni :bool = True, protocol :int = None) -> SSL.Connection:
        ctx = ssl.SSLContext() if protocol is None else ssl.SSLContext(protocol=protocol)
        ctx.verify_mode = ssl.CERT_NONE
        ctx.check_hostname = False
        if sock is None:
            sock = self.prepare_socket()
        conn = SSL.Connection(context, ctx.wrap_socket(sock, do_handshake_on_connect=False, server_hostname=self.host))
        conn.connect((self.host, self.port))
        conn.set_connect_state()
        if all([ssl.HAS_SNI, use_sni]):
            conn.set_tlsext_host_name(idna.encode(self.host))
        conn.setblocking(1)
        return conn

    @staticmethod
    def decode_http2_frame(frame_data) -> Frame:
        try:
            f, length = Frame.parse_frame_header(frame_data[:9])
            f.parse_body(memoryview(frame_data[9:9 + length]))
        except InvalidFrameError:
            f = None
        return f

    def _http2(self, conn :SSL.Connection, request :Frame, response_wait :int = 3) -> Frame:
        logger.info(f'HTTP/2 Request:\n{request}')
        try:
            conn.sendall(b'PRI * HTTP/2.0\r\n\r\nSM\r\n\r\n')
            conn.sendall(request.serialize())
        except SSL.WantReadError:
            sleep(0.5)
            return self._http2(conn, request, response_wait)
        except SSL.Error as err:
            if 'protocol is shutdown' not in str(err):
                logger.warning(err, exc_info=True)
        except SSL.ZeroReturnError:
            pass
        resp = None
        start = datetime.utcnow()
        frame_data = b''
        while True:
            try:
                if Transport.is_connection_closed(conn) or (datetime.utcnow()-start).seconds >= response_wait: break
                data = conn.recv(1024)
                if not data: break
                frame_data += data
            except (SSL.ZeroReturnError):
                break
            except Exception as ex:
                logger.warning(ex, exc_info=True)
                break
        try:
            resp = Transport.decode_http2_frame(frame_data)
            logger.info(f'Response data:\n{resp}')
            conn.shutdown()
        except Exception as ex:
            logger.warning(ex, exc_info=True)
        finally:
            conn.close()
        return resp

    def test_h2c(self, method :str = 'GET', uri_path :str = '/', response_wait :int = 3, include_settings :bool = True) -> Frame:
        logger.info(f'Testing HTTP/2 clear text')
        def _read(s :socket):
            resp = b''
            while not resp.endswith(b"\r\n\r\n"):
                try:
                    data = s.recv(1)
                except Exception:
                    break
                if not data: break
                resp += data
            return resp

        preamble = b'PRI * HTTP/2.0\r\n\r\nSM\r\n\r\n'
        req_parts = [
            f"{method} {uri_path} HTTP/1.1",
            f"Host: {self.host}",
            "Accept: */*",
            "Connection: Upgrade",
        ]
        if include_settings:
            settings_frame = SettingsFrame(0, settings={
                SettingsFrame.HEADER_TABLE_SIZE: 4096,
                SettingsFrame.INITIAL_WINDOW_SIZE: 2 ** 16 - 1,
                SettingsFrame.MAX_FRAME_SIZE: 2 ** 14,
            })
            http2_settings = urlsafe_b64encode(settings_frame.serialize_body()).rstrip(b'=')
            req_parts.append(f"HTTP2-Settings: {http2_settings.decode()}")
            req_parts.append("Upgrade: h2c, HTTP2-Settings")
        else:
            req_parts.append("Upgrade: h2c")
        request = '\r\n'.join(req_parts) + '\r\n\r\n'
        self.http2_cleartext_support = False
        sock = socket(AF_INET, SOCK_STREAM)
        sock.settimeout(response_wait)
        try:
            sock.connect((self.host, 80 if self.port == 443 else self.port))            
            sock.sendall(preamble)
            logger.info(f'HTTP/2 clear text Request,\n{request}')
            sock.sendall(request.encode())
            if include_settings:
                logger.info(f'HTTP2-Settings: {Transport.decode_http2_frame(settings_frame.serialize())}')
                sock.sendall(settings_frame.serialize())
        except Exception as ex:
            logger.warning(ex, exc_info=True)
        
        response = _read(sock)
        logger.info(f'HTTP/2 clear text Response,\n{response.decode()}')
        if b'\r\n\r\n' in response:
            headers, _ = response.split(b'\r\n\r\n', 1)
            split_headers = headers.split()
            if split_headers[1] == b'101':
                self.http2_cleartext_support = True
        sock.close()

    def test_http2(self, uri_path :str = '/', response_wait :int = 3):
        if not any([ssl.HAS_ALPN, ssl.HAS_NPN]):
            return
        if not isinstance(response_wait, int):
            raise TypeError(f'response_wait {type(response_wait)}, int supported')
        if not isinstance(uri_path, str):
            raise AttributeError('uri_path not supported')
        if not uri_path.startswith('/'):
            uri_path = f'/{uri_path}'
        ctx = self.prepare_context()
        ctx.set_alpn_protos([b'h2', b'http/1.1'])
        # rfc7540 section 9.2
        ctx.set_options(ssl.OP_NO_COMPRESSION | ssl.OP_NO_SSLv2 | ssl.OP_NO_SSLv3 | ssl.OP_NO_TLSv1 | ssl.OP_NO_TLSv1_1 | ssl.OP_NO_TLSv1_3)
        ctx.set_cipher_list(b"ECDHE+AESGCM:ECDHE+CHACHA20:DHE+AESGCM:DHE+CHACHA20")
        conn = self.prepare_connection(context=ctx, sock=self.prepare_socket(response_wait))
        conn.settimeout(response_wait)
        util.do_handshake(conn)
        request = Transport.decode_http2_frame(b'\x00\x00\x13\x00\x09\x00\x00\x00\x01\x0Ahello' + b'\0' * 10)
        resp = self._http2(conn, request, response_wait)
        if isinstance(resp, Frame):
            self.http2_support = True
            self.http2_response_frame = str(resp)

    def test_highest_tls_version(self, version, response_wait :int = 3):
        logger.warning(DeprecationWarning('Transport.test_highest_tls_version() was deprecated in version 0.4.4 and will be removed in version 0.5.0'), exc_info=True)
        self.test_tls_version(min_tls_version=version, response_wait=response_wait)

    def test_tls_version(self, min_tls_version :int = None, max_tls_version :int = None, use_sni :bool = False, response_wait :int = 3) -> str:
        if min_tls_version is not None and not isinstance(min_tls_version, int):
            raise TypeError(f'min_tls_version {type(min_tls_version)}, int supported')
        if max_tls_version is not None and not isinstance(max_tls_version, int):
            raise TypeError(f'max_tls_version {type(max_tls_version)}, int supported')
        if not isinstance(response_wait, int):
            raise TypeError(f'response_wait {type(response_wait)}, int supported')
        protocol = None
        ctx = self.prepare_context()
        ctx.set_options(_util.lib.SSL_OP_TLS_ROLLBACK_BUG)
        if min_tls_version is not None:
            logger.info(f'min protocol {util.OPENSSL_VERSION_LOOKUP[min_tls_version]}')
            ctx.set_min_proto_version(min_tls_version)
        if max_tls_version is not None:
            logger.info(f'max protocol {util.OPENSSL_VERSION_LOOKUP[max_tls_version]}')
            ctx.set_max_proto_version(max_tls_version)
        conn = self.prepare_connection(context=ctx, sock=self.prepare_socket(timeout=response_wait), use_sni=use_sni)
        conn.settimeout(response_wait)
        util.do_handshake(conn)
        protocol = conn.get_protocol_version_name()
        logger.info(f'Negotiated {protocol}')
        conn.shutdown()
        conn.close()
        return protocol

    def connect(self, tls_version :int, use_sni :bool = False, protocol :str = None):
        logger.info(f'Trying TLS version {util.OPENSSL_VERSION_LOOKUP[tls_version]}')
        ctx = self.prepare_context()
        ctx.set_verify(getattr(SSL, Transport.default_connect_verify_mode), self._verifier)
        ctx.set_max_proto_version(tls_version)
        ctx.set_ocsp_client_callback(self._ocsp_handler)
        ctx.set_options(_util.lib.SSL_OP_ALLOW_UNSAFE_LEGACY_RENEGOTIATION | _util.lib.SSL_OP_LEGACY_SERVER_CONNECT)
        conn = SSL.Connection(context=ctx, socket=self.prepare_socket())
        conn.request_ocsp()
        if all([use_sni, ssl.HAS_SNI]):
            logger.info('using SNI')
            conn.set_tlsext_host_name(idna.encode(self.host))
        try:
            conn.connect((self.host, self.port))
            conn.set_connect_state()
            conn.setblocking(1)
            util.do_handshake(conn)
            self.negotiated_cipher = conn.get_cipher_name()
            self.negotiated_cipher_bits = conn.get_cipher_bits()
            negotiated_protocol = conn.get_protocol_version_name()
            self.negotiated_protocol = f'{negotiated_protocol} ({hex(util.PROTOCOL_VERSION[negotiated_protocol])})'
            self.offered_tls_versions.append(self.negotiated_protocol)
            self.session_cache_mode = util.SESSION_CACHE_MODE[native_openssl.SSL_CTX_get_session_cache_mode(conn._context._context)]
            self.session_tickets = native_openssl.SSL_SESSION_has_ticket(conn.get_session()._session) == 1
            self.session_ticket_hints = native_openssl.SSL_SESSION_get_ticket_lifetime_hint(conn.get_session()._session) > 0
            self.peer_address, _ = conn.getpeername()
            self.ciphers = conn.get_cipher_list()
            if not isinstance(self.server_certificate, X509):
                self.server_certificate = conn.get_peer_certificate()
                for (_, cert) in enumerate(conn.get_peer_cert_chain()):
                    self.certificate_chain.append(cert)
                logger.debug(f'Peer cert chain length: {len(self.certificate_chain)}')
            if protocol is not None:
                logger.info(f'Trying protocol {protocol}')
                self._protocol_handler(conn, protocol)
            conn.shutdown()
        except SSL.Error as err:
            if not any(x in str(err) for x in ['no protocols available', 'alert protocol', 'shutdown while in init']):
                logger.warning(err, exc_info=True)
        except Exception as ex:
            logger.warning(ex, exc_info=True)
        finally:
            conn.close()

    def test_scsv(self, tls_version :int, use_sni :bool = True):
        """
        A connection of higher protocol indicates we have previously connected using the client fallback mechanism
        When a handshake is rejected (False) assume downgrade attacks were prevented
        """
        logger.info('Trying to derive SCSV')
        try:
            negotiated = self.test_tls_version(min_tls_version=tls_version, use_sni=use_sni)
            if negotiated:    
                self.preferred_protocol = f'{negotiated} ({hex(util.PROTOCOL_VERSION[negotiated])})'
            else:
                self.preferred_protocol = self.negotiated_protocol
            self.offered_tls_versions.append(self.preferred_protocol)
            self.tls_downgrade = negotiated is not None
        except Exception:
            self.tls_downgrade = False

    def test_tls_all_versions(self, use_sni :bool = True):
        logger.info('Testing all TLS versions')
        for ver_name, tls_version in util.PROTOCOL_VERSION.items():
            ver_display_name = f'{ver_name} ({hex(tls_version)})'
            if ver_name in FAKE_PROTOCOLS or ver_display_name in self.offered_tls_versions:
                continue
            try:
                negotiated = self.test_tls_version(max_tls_version=tls_version, use_sni=use_sni)
                supported = negotiated == ver_name
            except Exception:
                supported = False
            if supported is True:
                self.offered_tls_versions.append(ver_display_name)

    def test_tls_long_handshake_intolerance(self, version :int = None) -> bool:
        """
        If the Client Hello messages longer than 255 bytes and the connection fails
        Using ALL_CIPHERS is 3458 bytes so our Client Hello will be sufficiently long
        """
        native_ssl_openssl_version_map = {
            769: ssl.PROTOCOL_TLSv1,
            770: ssl.PROTOCOL_TLSv1_1,
            771: ssl.PROTOCOL_TLSv1_2,
            772: ssl.PROTOCOL_TLSv1_2,
        }
        protocol = native_ssl_openssl_version_map[version]
        ctx = ssl.SSLContext(protocol)
        sock = socket(AF_INET, SOCK_STREAM)
        context = SSL.Context(method=getattr(SSL, self.default_connect_method))
        if version == SSL.TLS1_3_VERSION:
            ctx.options |= ssl.OP_NO_SSLv2
            ctx.options |= ssl.OP_NO_SSLv3
            ctx.options |= ssl.OP_NO_TLSv1
            ctx.options |= ssl.OP_NO_TLSv1_1
            ctx.options |= ssl.OP_NO_TLSv1_2
            ctx.options |= ssl.OP_NO_COMPRESSION
            context.set_options(SSL.OP_NO_SSLv2 | SSL.OP_NO_SSLv3 | SSL.OP_NO_TLSv1 | SSL.OP_NO_TLSv1_1 | SSL.OP_NO_TLSv1_2 | SSL.OP_NO_COMPRESSION)
        if version == SSL.TLS1_2_VERSION:
            ctx.options |= ssl.OP_NO_SSLv2
            ctx.options |= ssl.OP_NO_SSLv3
            ctx.options |= ssl.OP_NO_TLSv1
            ctx.options |= ssl.OP_NO_TLSv1_1
            ctx.options |= ssl.OP_NO_TLSv1_3
            ctx.options |= ssl.OP_NO_COMPRESSION
            context.set_options(SSL.OP_NO_SSLv2 | SSL.OP_NO_SSLv3 | SSL.OP_NO_TLSv1 | SSL.OP_NO_TLSv1_1 | SSL.OP_NO_TLSv1_3 | SSL.OP_NO_COMPRESSION)
        if version == SSL.TLS1_1_VERSION:
            ctx.options |= ssl.OP_NO_SSLv2
            ctx.options |= ssl.OP_NO_SSLv3
            ctx.options |= ssl.OP_NO_TLSv1
            ctx.options |= ssl.OP_NO_TLSv1_2
            ctx.options |= ssl.OP_NO_TLSv1_3
            ctx.options |= ssl.OP_NO_COMPRESSION
            context.set_options(SSL.OP_NO_SSLv2 | SSL.OP_NO_SSLv3 | SSL.OP_NO_TLSv1 | SSL.OP_NO_TLSv1_2 | SSL.OP_NO_TLSv1_3 | SSL.OP_NO_COMPRESSION)
        if version == SSL.TLS1_VERSION:
            ctx.options |= ssl.OP_NO_SSLv2
            ctx.options |= ssl.OP_NO_SSLv3
            ctx.options |= ssl.OP_NO_TLSv1_1
            ctx.options |= ssl.OP_NO_TLSv1_2
            ctx.options |= ssl.OP_NO_TLSv1_3
            ctx.options |= ssl.OP_NO_COMPRESSION
            context.set_options(SSL.OP_NO_SSLv2 | SSL.OP_NO_SSLv3 | SSL.OP_NO_TLSv1_1 | SSL.OP_NO_TLSv1_2 | SSL.OP_NO_TLSv1_3 | SSL.OP_NO_COMPRESSION)

        sock.settimeout(1)
        ctx.verify_mode = ssl.CERT_REQUIRED
        context.verify_mode = ssl.CERT_REQUIRED
        context.load_verify_locations(cafile=where())
        for cafile in self.cafiles:
            context.load_verify_locations(cafile=cafile)
        if self.client_certificate_expected and isinstance(self.client_pem_path, str):
            context.use_certificate_file(certfile=self.client_pem_path, filetype=FILETYPE_PEM)
        ctx.set_ciphers(util.ALL_CIPHERS)
        context.set_cipher_list(util.ALL_CIPHERS.encode())
        context.set_min_proto_version(version)
        context.set_max_proto_version(version)
        conn = SSL.Connection(context, ctx.wrap_socket(sock, do_handshake_on_connect=False, server_hostname=self.host))
        negotiated_cipher = None
        try:
            conn.connect((self.host, self.port))
            conn.set_connect_state()
            if ssl.HAS_SNI:
                conn.set_tlsext_host_name(idna.encode(self.host))
            conn.setblocking(1)
            util.do_handshake(conn)
            negotiated_cipher = conn.get_cipher_name()
            conn.shutdown()
        except Exception as ex:
            logger.debug(ex, exc_info=True)
        finally:
            conn.close()
        return negotiated_cipher is None

    def test_tls_version_interference(self):
        """
        A rejected connection (typically the oldest or latest version, currently 1.3)
        when no mutual accepted TLS version can be negotiates is known as tls interference
        """
        for check_interference in ['TLSv1.3 (0x304)', 'TLSv1.2 (0x303)', 'TLSv1.1 (0x302)', 'TLSv1 (0x301)', 'SSLv3 (0x300)', 'SSLv2 (0x2ff)']:
            if check_interference not in self.offered_tls_versions:
                self.tls_version_interference_versions.append(check_interference)
        self.tls_version_interference = len(self.tls_version_interference_versions) > 0

    def test_tls_version_intolerance(self, use_sni :bool = True):
        """
        Protocol not understood by the server, the server should negotiate the highest protocol it knows
        A rejected connection indicates TLS version intolerance, and is not rfc5246 or rfc8446 compliant
        """
        logger.info('Trying to derive TLS version intolerance')
        for fake_proto in FAKE_PROTOCOLS:
            fake_ver = util.PROTOCOL_VERSION[fake_proto]
            try:
                intolerance = self.test_tls_version(min_tls_version=fake_ver, use_sni=use_sni) is None
                if intolerance:
                    self.tls_version_intolerance_versions.append(f'{fake_proto} ({hex(fake_ver)})')
            except Exception:
                self.tls_version_intolerance_versions.append(f'{fake_proto} ({hex(fake_ver)})')
        self.tls_version_intolerance = len(self.tls_version_intolerance_versions) > 0

    def connect_least_secure(self, cafiles :list = None, use_sni :bool = False, progress_bar :callable = lambda *args: None) -> bool:
        if not isinstance(self.port, int):
            raise TypeError(f"provided an invalid type {type(self.port)} for port, expected int")
        if validators.domain(self.host) is not True:
            raise ValueError(f"provided an invalid domain {self.host}")
        if cafiles is not None:
            if not isinstance(cafiles, list):
                raise TypeError(f"provided an invalid type {type(cafiles)} for cafiles, expected list")
            valid_cafiles = util.filter_valid_files_urls(cafiles)
            if valid_cafiles is False:
                raise AttributeError('cafiles was provided but is not a valid URLs or files do not exist')
            if isinstance(valid_cafiles, list): self.cafiles = valid_cafiles

        tls_versions = [SSL.SSL3_VERSION, SSL.TLS1_VERSION, SSL.TLS1_1_VERSION, SSL.TLS1_2_VERSION, SSL.TLS1_3_VERSION]
        for index, version in enumerate(tls_versions):
            self.connect(tls_version=version, use_sni=use_sni) # Skip HTTP testing until negotiated
            progress_bar()
            if not isinstance(self.server_certificate, X509):
                continue
            progress_bar(5)

            if all([use_sni, ssl.HAS_SNI]):
                self.sni_support = True
            for protocol in ['HTTP/1.0', 'HTTP/1.1']:
                self.connect(tls_version=version, use_sni=use_sni, protocol=protocol)
                progress_bar()

            self.test_h2c(response_wait=3)
            progress_bar()

            self.test_http2(response_wait=3)
            progress_bar()

            self.long_handshake_intolerance = self.test_tls_long_handshake_intolerance(version)
            progress_bar()

            if version == SSL.TLS1_3_VERSION:
                # Already the highest TLS protocol, no downgrade possible
                self.tls_downgrade = False
                # server can only prefer this too
                self.preferred_protocol = util.OPENSSL_VERSION_LOOKUP[version]
                self.test_tls_version_interference()
                progress_bar(12)
                self.test_tls_version_intolerance(use_sni) # sourcery skip: extract-duplicate-method
                progress_bar()
                return True

            self.test_scsv(tls_versions[index+1], use_sni)
            progress_bar()
            
            self.test_tls_all_versions(use_sni)
            progress_bar()

            self.test_tls_version_interference()
            progress_bar()

            self.test_tls_version_intolerance(use_sni)
            progress_bar()

            return True

        return False

    @staticmethod
    def is_connection_closed(conn: SSL.Connection, counter :int = 0, max_retries :int = 5) -> bool:
        try:
            data = conn.recv(1, MSG_PEEK)
            if len(data) == 0:
                return True
        except SSL.WantReadError:
            if counter >= max_retries: return True
            sleep(0.5)
            return Transport.is_connection_closed(conn, counter+1)
        except BlockingIOError:
            return False
        except SSL.ZeroReturnError:
            return True
        except ConnectionResetError:
            return True
        except SSL.SysCallError:
            return True
        except Exception as ex:
            logger.exception(ex)
            return False
        return False

    @staticmethod
    def parse_header(head :str) -> dict:
        ret = {'headers': {}, 'response_code': 0, 'response_status': '', 'protocol': ''}
        i = 0
        for line in head.splitlines():
            i += 1
            if len(line) == 0:
                continue
            if i == 1 and len(line.split(' ')) >= 2:
                ret['protocol'], ret['response_code'], *extra = line.split(' ')
                ret['response_status'] = ' '.join(extra)
            else:
                header = line.split(':')[0].lower()
                value = ':'.join(line.split(':')[1:]).strip()
                if header in ret['headers']:
                    prev = ret['headers'][header].split(', ')
                    ret['headers'][header] = ', '.join([value] + prev)
                else:
                    ret['headers'][header] = value
        return ret



__module__ = 'tlsverify.pci'
__version__ = '3.2.1'

WEAK_KEY_SIZE = {
    'RSA': 2048,
    'DSA': 2048,
    'EC': 256,
    'DH': 256,
}
WEAK_CIPHER_BITS = 128
PCIDSS_NON_COMPLIANCE_WEAK_KEY_RSA = f'PCI DSS version {__version__} compliance requires RSA public key bits to greater than {WEAK_KEY_SIZE["RSA"]}'
PCIDSS_NON_COMPLIANCE_WEAK_KEY_DSA = f'PCI DSS version {__version__} compliance requires DSA public key bits to greater than {WEAK_KEY_SIZE["DSA"]}'
PCIDSS_NON_COMPLIANCE_WEAK_KEY_EC = f'PCI DSS version {__version__} compliance requires Elliptic curve public key bits to greater than {WEAK_KEY_SIZE["EC"]}'
PCIDSS_NON_COMPLIANCE_WEAK_KEY_DH = f'PCI DSS version {__version__} compliance requires Diffie-Hellman exchange public key bits to greater than {WEAK_KEY_SIZE["DH"]}'
PCIDSS_NON_COMPLIANCE_WEAK_PROTOCOL = f'PCI DSS version {__version__} compliance requires deprecated and known weak TLS protocols are not supported'
PCIDSS_NON_COMPLIANCE_CIPHER = f'PCI DSS version {__version__} compliance requires cipher bits to greater than {WEAK_CIPHER_BITS}, not be of Anonymous key exchange suites, Exchange ciphers, or other known weak ciphers as thy are discovered'
PCIDSS_NON_COMPLIANCE_KNOWN_VULNERABILITIES = f'PCI DSS version {__version__} compliance requires that no known vulnerabilities are present'
PCIDSS_NON_COMPLIANCE_CA_TRUST = f'PCI DSS version {__version__} compliance requires a complete Certificate chain with verified trust anchor'
PCIDSS_NON_COMPLIANCE_WEAK_ALGORITHMS = f'PCI DSS version {__version__} compliance requires deprecated and known weak algorithms are not supported'
VALIDATION_CA_TRUST = 'pci_ca_trust'
VALIDATION_WEAK_KEY = 'pci_weak_key'
VALIDATION_WEAK_CIPHER = 'pci_weak_cipher'
VALIDATION_WEAK_PROTOCOL = 'pci_weak_protocol'
VALIDATION_DEPRECATED_ALGO = 'pci_deprecated_algo'
VALIDATION_KNOWN_VULN_COMPRESSION = 'pci_vuln_compression'
VALIDATION_KNOWN_VULN_RENEGOTIATION = 'pci_vuln_renegotiation'
VALIDATION_KNOWN_VULN_SESSION_RESUMPTION = 'pci_vuln_session_resumption'



import json
import hashlib
import logging
from os import path
from base64 import b64encode
from datetime import datetime
from pathlib import Path
import asn1crypto
from ssl import PEM_cert_to_DER_cert
from OpenSSL import SSL
from OpenSSL.crypto import X509,  X509Name, dump_privatekey, dump_certificate, load_certificate, FILETYPE_PEM, FILETYPE_ASN1, FILETYPE_TEXT, TYPE_RSA, TYPE_DSA, TYPE_DH, TYPE_EC
from cryptography.x509 import Certificate, extensions, PolicyInformation
from certvalidator.errors import PathValidationError, RevokedError, InvalidCertificateError, PathBuildingError
from dns.rrset import RRset
from tldextract import TLDExtract
from tlstrust import TrustStore, context
from tlstrust.stores.ccadb import __version__ as ccadb_version
from tlstrust.stores.java import __version__ as java_version
from tlstrust.stores.certifi import __version__ as certifi_version
from tlstrust.stores.linux import __version__ as linux_version
from tlstrust.stores.android_7 import __description__ as android7_version
from tlstrust.stores.android_8 import __description__ as android8_version
from tlstrust.stores.android_9 import __description__ as android9_version
from tlstrust.stores.android_10 import __description__ as android10_version
from tlstrust.stores.android_11 import __description__ as android11_version
from tlstrust.stores.android_12 import __description__ as android12_version
from . import exceptions, util, pci, fips, nist
from .transport import Transport
from .metadata import Metadata

__module__ = 'tlsverify.validator'
logger = logging.getLogger(__name__)

VALIDATION_CLIENT_AUTHENTICATION = 'client_authentication'
VALIDATION_CLIENT_AUTH_USAGE = 'client_certificate_permits_authentication_usage'
VALIDATION_NOT_EXPIRED = 'not_expired'
VALIDATION_ISSUED_PAST_TENSE = 'issued_past_tense'
VALIDATION_SUBJECT_CN_DEFINED = 'common_name_defined'
VALIDATION_SUBJECT_CN_VALID = 'common_name_valid'
VALIDATION_MATCH_HOSTNAME = 'match_hostname'
VALIDATION_NOT_SELF_SIGNED = 'not_self_signed'
VALIDATION_WEAK_SIG_ALGO = 'avoid_known_weak_signature_algorithm'
VALIDATION_WEAK_KEYS = 'avoid_known_weak_keys'
VALIDATION_DEPRECATED_TLS_PROTOCOLS = 'avoid_deprecated_protocols'
VALIDATION_DEPRECATED_DNSSEC_ALGO = 'avoid_deprecated_dnssec_algorithms'
VALIDATION_BASIC_CONSTRAINTS_CA = 'basic_constraints_ca'
VALIDATION_VALID_TLS_USAGE = 'certificate_valid_tls_usage'
VALIDATION_REVOCATION = 'not_revoked'
VALIDATION_ROOT_CA_TRUST = 'trusted_ca'
VALIDATION_VALID_DNSSEC = 'valid_dnssec'
VALIDATION_VALID_CAA = 'valid_caa'
VALIDATION_OCSP_STAPLE_SATISFIED = 'ocsp_staple_satisfied'
VALIDATION_OCSP_MUST_STAPLE_SATISFIED = 'ocsp_must_staple_satisfied'

class Validator:
    _pem :bytes
    _der :bytes
    x509 :X509
    certificate :Certificate
    tmp_path_prefix :str
    metadata :Metadata
    validation_checks :dict
    certificate_verify_messages :list

    def __init__(self, tmp_path_prefix :str = '/tmp') -> None:
        if not isinstance(tmp_path_prefix, str):
            raise TypeError(f'tmp_path_prefix of type {type(tmp_path_prefix)} not supported, str expected')
        tmp_path = Path(tmp_path_prefix)
        if not tmp_path.is_dir():
            raise AttributeError(f'tmp_path_prefix {tmp_path_prefix} is not a directory')
        self.tmp_path_prefix = tmp_path_prefix
        self.validation_checks = {}
        self.certificate_verify_messages = []
        self._pem = None
        self._der = None
        self.x509 = None
        self.certificate = None
        self.metadata = None
    
    @property
    def certificate_valid(self):
        validations = list(self.validation_checks.values())
        if not validations:
            return False
        return all(validations)

    def init_der(self, der :bytes):
        self._der = der
        logger.debug('added ASN1/DER bytes')
        self.x509 = load_certificate(FILETYPE_ASN1, der)
        logger.debug('added pyOpenSSL x509')
        self._pem = dump_certificate(FILETYPE_PEM, self.x509)
        logger.debug('added PEM bytes')
        self.certificate = self.x509.to_cryptography()
        logger.debug('added lib cryptography object')
        self.extract_x509_metadata(self.x509)

    def init_pem(self, pem :bytes):
        self._pem = pem
        logger.debug('added PEM bytes')
        self.x509 = load_certificate(FILETYPE_PEM, pem)
        logger.debug('added pyOpenSSL x509')
        self._der = PEM_cert_to_DER_cert(self._pem.decode())
        logger.debug('added ASN1/DER bytes')
        self.certificate = self.x509.to_cryptography()
        logger.debug('added lib cryptography object')
        self.extract_x509_metadata(self.x509)

    def init_x509(self, x509 :X509):
        self.x509 = x509
        logger.debug('added pyOpenSSL x509')
        self._pem = dump_certificate(FILETYPE_PEM, x509)
        logger.debug('added PEM bytes')
        self._der = PEM_cert_to_DER_cert(self._pem.decode())
        logger.debug('added ASN1/DER bytes')
        self.certificate = x509.to_cryptography()
        logger.debug('added lib cryptography object')
        self.extract_x509_metadata(self.x509)

    def cert_to_text(self) -> str:
        logger.debug('dump_certificate x509 FILETYPE_TEXT')
        return dump_certificate(FILETYPE_TEXT, self.x509).decode()

    def extract_x509_metadata(self, x509 :X509):
        if not hasattr(self, 'metadata') or not isinstance(self.metadata, Metadata):
            self.metadata = Metadata()
        self.metadata.certificate_version = x509.get_version()
        self.metadata.certificate_extensions = util.get_certificate_extensions(self.certificate)
        self.metadata.certificate_private_key_pem = None
        public_key = x509.get_pubkey()
        self._extract_public_key_info(public_key.type(), public_key, x509)
        self.metadata.certificate_serial_number_decimal = x509.get_serial_number()
        self.metadata.certificate_serial_number = util.convert_decimal_to_serial_bytes(x509.get_serial_number())
        self.metadata.certificate_serial_number_hex = '{0:#0{1}x}'.format(x509.get_serial_number(), 4)
        subject = x509.get_subject()
        self.metadata.certificate_subject = "".join("/{0:s}={1:s}".format(name.decode(), value.decode()) for name, value in subject.get_components())
        issuer: X509Name = x509.get_issuer()
        self.metadata.certificate_issuer = issuer.commonName
        self.metadata.certificate_issuer_country = issuer.countryName
        self.metadata.certificate_signature_algorithm = x509.get_signature_algorithm().decode('ascii')
        self.metadata.certificate_pin_sha256 = b64encode(hashlib.sha256(asn1crypto.x509.Certificate.load(self._der).public_key.dump()).digest()).decode()
        self.metadata.certificate_sha256_fingerprint = hashlib.sha256(self._der).hexdigest()
        self.metadata.certificate_sha1_fingerprint = hashlib.sha1(self._der).hexdigest()
        self.metadata.certificate_md5_fingerprint = hashlib.md5(self._der).hexdigest()
        self.metadata.certificate_san = util.get_san(self.certificate)
        not_before = datetime.strptime(x509.get_notBefore().decode('ascii'), util.X509_DATE_FMT)
        not_after = datetime.strptime(x509.get_notAfter().decode('ascii'), util.X509_DATE_FMT)
        self.metadata.certificate_not_before = not_before.isoformat()
        self.metadata.certificate_not_after = not_after.isoformat()
        self.metadata.certificate_expired = x509.has_expired()
        self.metadata.certificate_common_name = util.extract_from_subject(self.certificate)
        self.metadata.certificate_subject_key_identifier, self.metadata.certificate_authority_key_identifier = util.get_ski_aki(self.certificate)
        if self.metadata.revocation_ocsp_must_staple is not True:
            for ext in self.metadata.certificate_extensions:
                if ext['name'] == 'TLSFeature' and 'rfc6066' in ext['features']:
                    self.metadata.revocation_ocsp_must_staple = True
        policies = []
        try:
            policies = self.certificate.extensions.get_extension_for_class(extensions.CertificatePolicies).value._policies
        except extensions.ExtensionNotFound:
            pass
        for policy in policies:
            if not isinstance(policy, PolicyInformation): continue
            if policy.policy_identifier._dotted_string in util.VALIDATION_OID.keys():
                self.metadata.certificate_validation_type = util.VALIDATION_TYPES[util.VALIDATION_OID[policy.policy_identifier._dotted_string]]

    def _extract_public_key_info(self, key_type, public_key, x509):
        if key_type in [TYPE_RSA, TYPE_DSA]:
            self.metadata.certificate_public_key_type = 'RSA' if TYPE_RSA == key_type else 'DSA'
            self.metadata.certificate_private_key_pem = dump_privatekey(FILETYPE_PEM, public_key).decode()
            self.metadata.certificate_public_key_exponent = x509.to_cryptography().public_key().public_numbers().e
        if key_type in [TYPE_DH, TYPE_EC]:
            self.metadata.certificate_public_key_type = 'EC' if key_type == TYPE_EC else 'DH'
            self.metadata.certificate_public_key_curve = x509.to_cryptography().public_key().curve.name
        self.metadata.certificate_public_key_size = public_key.bits()

    def pcidss_compliant(self) -> bool:
        logger.debug('PCI DSS compliance validations')
        self.validation_checks[pci.VALIDATION_WEAK_KEY] = self.metadata.certificate_public_key_size >= pci.WEAK_KEY_SIZE[self.metadata.certificate_public_key_type]
        if self.validation_checks[pci.VALIDATION_WEAK_KEY] is False:
            if self.metadata.certificate_public_key_type == 'RSA':
                self.certificate_verify_messages.append(pci.PCIDSS_NON_COMPLIANCE_WEAK_KEY_RSA)
            if self.metadata.certificate_public_key_type == 'DSA':
                self.certificate_verify_messages.append(pci.PCIDSS_NON_COMPLIANCE_WEAK_KEY_DSA)
            if self.metadata.certificate_public_key_type == 'EC':
                self.certificate_verify_messages.append(pci.PCIDSS_NON_COMPLIANCE_WEAK_KEY_EC)
            if self.metadata.certificate_public_key_type == 'DH':
                self.certificate_verify_messages.append(pci.PCIDSS_NON_COMPLIANCE_WEAK_KEY_DH)

    def fips_compliant(self) -> bool:
        logger.debug('FIPS compliance validations')
        self.validation_checks[fips.VALIDATION_WEAK_KEY] = self.metadata.certificate_public_key_size >= fips.WEAK_KEY_SIZE[self.metadata.certificate_public_key_type]
        if self.validation_checks[fips.VALIDATION_WEAK_KEY] is False:
            if self.metadata.certificate_public_key_type == 'RSA':
                self.certificate_verify_messages.append(fips.FIPS_NON_COMPLIANCE_WEAK_KEY_RSA)
            if self.metadata.certificate_public_key_type == 'DSA':
                self.certificate_verify_messages.append(fips.FIPS_NON_COMPLIANCE_WEAK_KEY_DSA)
            if self.metadata.certificate_public_key_type == 'EC':
                self.certificate_verify_messages.append(fips.FIPS_NON_COMPLIANCE_WEAK_KEY_EC)
            if self.metadata.certificate_public_key_type == 'DH':
                self.certificate_verify_messages.append(fips.FIPS_NON_COMPLIANCE_WEAK_KEY_DH)

    def nist_compliant(self) -> bool:
        logger.debug('NIST compliance validations')
        self.validation_checks[nist.VALIDATION_WEAK_KEY] = self.metadata.certificate_public_key_size >= nist.WEAK_KEY_SIZE[self.metadata.certificate_public_key_type]
        if self.validation_checks[nist.VALIDATION_WEAK_KEY] is False:
            if self.metadata.certificate_public_key_type == 'RSA':
                self.certificate_verify_messages.append(nist.NIST_NON_COMPLIANCE_WEAK_KEY_RSA)
            if self.metadata.certificate_public_key_type == 'DSA':
                self.certificate_verify_messages.append(nist.NIST_NON_COMPLIANCE_WEAK_KEY_DSA)
            if self.metadata.certificate_public_key_type == 'EC':
                self.certificate_verify_messages.append(nist.NIST_NON_COMPLIANCE_WEAK_KEY_EC)
            if self.metadata.certificate_public_key_type == 'DH':
                self.certificate_verify_messages.append(nist.NIST_NON_COMPLIANCE_WEAK_KEY_DH)

    def verify(self) -> bool:
        logger.debug('Common certificate validations')
        not_after = datetime.fromisoformat(self.metadata.certificate_not_after)
        not_before = datetime.fromisoformat(self.metadata.certificate_not_before)
        self.validation_checks[VALIDATION_NOT_EXPIRED] = not_after > datetime.utcnow()
        if self.validation_checks[VALIDATION_NOT_EXPIRED] is False:
            self.certificate_verify_messages.append(util.date_diff(not_after))
        self.validation_checks[VALIDATION_ISSUED_PAST_TENSE] = not_before < datetime.utcnow()
        if self.validation_checks[VALIDATION_ISSUED_PAST_TENSE] is False:
            self.certificate_verify_messages.append(f'Will only be valid for use in {(datetime.utcnow() - self.metadata.certificate_not_before).days} days')
        self.validation_checks[VALIDATION_SUBJECT_CN_DEFINED] = self.metadata.certificate_common_name is not None
        self.validation_checks[VALIDATION_WEAK_SIG_ALGO] = self.metadata.certificate_signature_algorithm not in util.KNOWN_WEAK_SIGNATURE_ALGORITHMS.keys()
        if self.validation_checks[VALIDATION_WEAK_SIG_ALGO] is False:
            self.certificate_verify_messages.append(util.KNOWN_WEAK_SIGNATURE_ALGORITHMS[self.metadata.certificate_signature_algorithm])
        self.validation_checks[VALIDATION_WEAK_KEYS] = self.metadata.certificate_public_key_type not in util.KNOWN_WEAK_KEYS.keys() or self.metadata.certificate_public_key_size > util.WEAK_KEY_SIZE[self.metadata.certificate_public_key_type]
        if self.validation_checks[VALIDATION_WEAK_KEYS] is False:
            self.certificate_verify_messages.append(util.KNOWN_WEAK_KEYS[self.metadata.certificate_public_key_type])
        logger.debug('Impersonation, C2, other detections')
        bad_cn = ['localhost', 'portswigger']
        bad_san = ['localhost', 'lvh.me']
        for cn in bad_cn:
            check = self.metadata.certificate_common_name
            if not check:
                continue
            if cn in check.lower():
                self.metadata.possible_phish_or_malicious = True
        subject_ou = util.extract_from_subject(self.certificate, 'organizationalUnitName')
        if subject_ou:
            bad_ou = ['charlesproxy', 'portswigger']
            for ou in bad_ou:
                if ou in subject_ou.lower():
                    self.metadata.possible_phish_or_malicious = True
        for bad in bad_san:
            for san in self.metadata.certificate_san:
                if bad in san:
                    self.metadata.possible_phish_or_malicious = True
        self.metadata.revocation_crlite = util.crlite_revoked(db_path=path.join(self.tmp_path_prefix, ".crlite_db"), pem=self._pem)
        if self.metadata.revocation_crlite:
            self.validation_checks[VALIDATION_REVOCATION] = False
        if self.metadata.certificate_private_key_pem and 'BEGIN PRIVATE KEY' in self.metadata.certificate_private_key_pem:
            self.certificate_verify_messages.append(exceptions.VALIDATION_ERROR_EXPOSED_PRIVATE_KEY)
        #TODO OneCRL

class RootCertValidator(Validator):
    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)

    def __repr__(self) -> str:
        certificate_verify_messages = '", "'.join(self.certificate_verify_messages)
        validation_checks = json.dumps(self.validation_checks)
        return f'<Validator(certificate_valid={self.certificate_valid}, ' +\
              f'certificate_verify_messages=["{certificate_verify_messages}"]", ' +\
              f'validation_checks={validation_checks}, ' +\
               'metadata=<tlsverify.metadata.Metadata>, ' +\
               'x509=<OpenSSL.crypto.X509>, ' +\
               '_pem=<bytes>, ' +\
               '_der=<bytes>, ' +\
               'certificate=<cryptography.x509.Certificate>)>'

    def pcidss_compliant(self) -> bool:
        super().pcidss_compliant()
        self.validation_checks[pci.VALIDATION_CA_TRUST] = all([self.metadata.trust_ccadb, self.metadata.trust_android, self.metadata.trust_certifi, self.metadata.trust_java, self.metadata.trust_linux])
        if self.validation_checks[pci.VALIDATION_CA_TRUST] is False:
            self.certificate_verify_messages.append(pci.PCIDSS_NON_COMPLIANCE_CA_TRUST)

    def fips_compliant(self) -> bool:
        super().fips_compliant()
        self.validation_checks[fips.VALIDATION_CA_TRUST] = all([self.metadata.trust_ccadb, self.metadata.trust_android, self.metadata.trust_certifi, self.metadata.trust_java, self.metadata.trust_linux])
        if self.validation_checks[fips.VALIDATION_CA_TRUST] is False:
            self.certificate_verify_messages.append(fips.FIPS_NON_COMPLIANCE_CA_TRUST)

    def nist_compliant(self) -> bool:
        super().nist_compliant()
        self.validation_checks[nist.VALIDATION_CA_TRUST] = all([self.metadata.trust_ccadb, self.metadata.trust_android, self.metadata.trust_certifi, self.metadata.trust_java, self.metadata.trust_linux])
        if self.validation_checks[nist.VALIDATION_CA_TRUST] is False:
            self.certificate_verify_messages.append(nist.NIST_NON_COMPLIANCE_CA_TRUST)

    def verify_trust(self, trust_store :TrustStore):
        DEFAULT_STATUS = 'No Root CA Certificate in the {platform} Trust Store'
        self.metadata.trust_ccadb_status = DEFAULT_STATUS.format(platform='CCADB')
        self.metadata.trust_android_status = DEFAULT_STATUS.format(platform='Android')
        self.metadata.trust_java_status = DEFAULT_STATUS.format(platform='Java')
        self.metadata.trust_linux_status = DEFAULT_STATUS.format(platform='Linux')
        self.metadata.trust_certifi_status = DEFAULT_STATUS.format(platform='Python')

        expired_text = ' EXPIRED'
        if trust_store.exists(context.SOURCE_CCADB):
            self.metadata.trust_ccadb_status = f'In Common CA Database {ccadb_version} (Mozilla, Microsoft, and Apple)'
            if trust_store.expired_in_store(context.SOURCE_CCADB):
                self.metadata.trust_ccadb_status += expired_text
        if trust_store.exists(context.SOURCE_JAVA):
            self.metadata.trust_java_status = f'In Java Root CA Trust Store {java_version}'
            if trust_store.expired_in_store(context.SOURCE_JAVA):
                self.metadata.trust_java_status += expired_text
        if trust_store.exists(context.SOURCE_LINUX):
            self.metadata.trust_linux_status = f'In Linux {linux_version} Root CA Trust Store'
            if trust_store.expired_in_store(context.SOURCE_LINUX):
                self.metadata.trust_linux_status += expired_text
        if trust_store.exists(context.SOURCE_CERTIFI):
            self.metadata.trust_certifi_status = f'In Python {certifi_version} Root CA Trust Store (Django, requests, urllib, and anything based from these)'
            if trust_store.expired_in_store(context.SOURCE_CERTIFI):
                self.metadata.trust_certifi_status += expired_text

        android_stores = []
        if trust_store.exists(context.PLATFORM_ANDROID7):
            android_status = android7_version
            if trust_store.expired_in_store(context.PLATFORM_ANDROID7):
                android_status += expired_text
            android_stores.append(android_status)
        if trust_store.exists(context.PLATFORM_ANDROID8):
            android_status = android8_version
            if trust_store.expired_in_store(context.PLATFORM_ANDROID8):
                android_status += expired_text
            android_stores.append(android_status)
        if trust_store.exists(context.PLATFORM_ANDROID9):
            android_status = android9_version
            if trust_store.expired_in_store(context.PLATFORM_ANDROID9):
                android_status += expired_text
            android_stores.append(android_status)
        if trust_store.exists(context.PLATFORM_ANDROID10):
            android_status = android10_version
            if trust_store.expired_in_store(context.PLATFORM_ANDROID10):
                android_status += expired_text
            android_stores.append(android_status)
        if trust_store.exists(context.PLATFORM_ANDROID11):
            android_status = android11_version
            if trust_store.expired_in_store(context.PLATFORM_ANDROID11):
                android_status += expired_text
            android_stores.append(android_status)
        if trust_store.exists(context.PLATFORM_ANDROID12):
            android_status = android12_version
            if trust_store.expired_in_store(context.PLATFORM_ANDROID12):
                android_status += expired_text
            android_stores.append(android_status)
        if android_stores:
            self.metadata.trust_android_status = "\n".join(android_stores)

        self.metadata.trust_ccadb = trust_store.ccadb
        self.metadata.trust_java = trust_store.java
        self.metadata.trust_android = all([trust_store.android7, trust_store.android8, trust_store.android9, trust_store.android10, trust_store.android11, trust_store.android12])
        self.metadata.trust_linux = trust_store.linux
        self.metadata.trust_certifi = trust_store.certifi

class PeerCertValidator(Validator):
    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)

    def __repr__(self) -> str:
        certificate_verify_messages = '", "'.join(self.certificate_verify_messages)
        validation_checks = json.dumps(self.validation_checks)
        return f'<Validator(certificate_valid={self.certificate_valid}, ' +\
              f'certificate_verify_messages=["{certificate_verify_messages}"]", ' +\
              f'validation_checks={validation_checks}, ' +\
               'metadata=<tlsverify.metadata.Metadata>, ' +\
               'x509=<OpenSSL.crypto.X509>, ' +\
               '_pem=<bytes>, ' +\
               '_der=<bytes>, ' +\
               'certificate=<cryptography.x509.Certificate>)>'

class CertValidator(Validator):
    _pem_certificate_chain :list
    peer_validations :list[PeerCertValidator]
    certificate_chain :list[X509]
    transport :Transport
    _root_certs :list

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)
        self.transport = None
        self.certificate_chain_valid = None
        self.certificate_chain_validation_result = None
        self._pem_certificate_chain = []
        self.peer_validations = []
        self.certificate_chain = []

    def __repr__(self) -> str:
        certificate_verify_messages = '", "'.join(self.certificate_verify_messages)
        validation_checks = json.dumps(self.validation_checks)
        return (
            f'<Validator(certificate_valid={self.certificate_valid}, '
            + f'certificate_chain_valid={self.certificate_chain_valid}, '
            + f'certificate_chain_validation_result={self.certificate_chain_validation_result}, '
            + f'certificate_verify_messages=["{certificate_verify_messages}"]", '
            + f'validation_checks={validation_checks}, '
            + 'metadata=<tlsverify.metadata.Metadata>, '
            + 'transport=<tlsverify.transport.Transport>, '
            + 'x509=<OpenSSL.crypto.X509>, '
            + '_pem=<bytes>, '
            + '_der=<bytes>, '
            + 'certificate=<cryptography.x509.Certificate>)>'
        )

    def mount(self, transport :Transport):
        if not isinstance(transport, Transport):
            raise TypeError(f"provided an invalid type {type(transport)} for transport, expected an instance of <tlsverify.transport.Transport>")
        self.transport = transport
        self.extract_transport_metadata(transport)
        if isinstance(transport.server_certificate, X509):
            self.init_x509(transport.server_certificate)
        self.parse_openssl_errors(transport.verifier_errors)

    def extract_transport_metadata(self, transport :Transport):
        if not hasattr(self, 'metadata') or not isinstance(self.metadata, Metadata):
            self.metadata = Metadata(
                host=transport.host,
                port=transport.port
            )
        self._pem_certificate_chain = util.convert_x509_to_PEM(transport.certificate_chain)
        self.certificate_chain = transport.certificate_chain
        self.metadata.peer_address = transport.peer_address
        self.metadata.revocation_ocsp_stapling = transport.ocsp_stapling
        self.metadata.revocation_ocsp_must_staple = transport.ocsp_must_staple
        self.metadata.revocation_ocsp_status = transport.ocsp_certificate_status
        self.metadata.revocation_ocsp_response = transport.ocsp_response_status
        self.metadata.revocation_ocsp_reason = transport.ocsp_revocation_reason
        self.metadata.revocation_ocsp_time = transport.ocsp_revocation_time
        self.metadata.offered_ciphers = transport.ciphers
        self.metadata.client_certificate_expected = transport.client_certificate_expected is True
        self.metadata.negotiated_cipher_bits = transport.negotiated_cipher_bits
        self.metadata.negotiated_cipher = transport.negotiated_cipher
        if self.metadata.negotiated_cipher and any(self.metadata.negotiated_cipher.startswith(c) for c in ['DHE', 'ECDHE']):
            self.metadata.forward_anonymity = True
        if self.metadata.negotiated_cipher not in util.NOT_KNOWN_WEAK_CIPHERS:
            self.metadata.weak_cipher = True
        if self.metadata.negotiated_cipher in util.STRONG_CIPHERS:
            self.metadata.strong_cipher = True
        self.metadata.negotiated_protocol = transport.negotiated_protocol
        self.metadata.sni_support = transport.sni_support
        self.metadata.tls_version_intolerance = transport.tls_version_intolerance
        self.metadata.tls_version_intolerance_versions = transport.tls_version_intolerance_versions
        self.metadata.tls_version_interference = transport.tls_version_interference
        self.metadata.tls_version_interference_versions = transport.tls_version_interference_versions
        self.metadata.tls_long_handshake_intolerance = transport.long_handshake_intolerance
        self.metadata.offered_tls_versions = list(set(transport.offered_tls_versions))
        self.metadata.session_resumption_caching = transport.session_cache_mode in ['session_resumption_both', 'session_resumption_caching']
        self.metadata.session_resumption_tickets = transport.session_tickets
        self.metadata.session_resumption_ticket_hint = transport.session_ticket_hints
        self.metadata.compression_support = self.header_exists(name='content-encoding', includes_value='gzip')
        self.metadata.client_renegotiation = transport.client_renegotiation
        self.metadata.scsv = transport.tls_downgrade is False # modern network tools such as F5 try to hide cause and blend in making scsv undetectable directly but a downgrade can still be observed
        self.metadata.preferred_protocol = transport.preferred_protocol
        self.metadata.http_hsts = self.header_exists(name='strict-transport-security', includes_value='max-age')
        self.metadata.http_expect_ct_report_uri = self.header_exists(name='expect-ct', includes_value='report-uri')
        self.metadata.http_xfo = self.header_exists(name='x-frame-options') or self.header_exists(name='content-security-policy', includes_value='frame-ancestors')
        self.metadata.http_nosniff = self.header_exists(name='x-content-type-options', includes_value='nosniff')
        self.metadata.http_csp = self.header_exists(name='content-security-policy')
        self.metadata.http_coep = self.header_exists(name='cross-origin-embedder-policy', includes_value='require-corp')
        self.metadata.http_coop = self.header_exists(name='cross-origin-opener-policy', includes_value='same-origin')
        self.metadata.http_corp = self.header_exists(name='cross-origin-resource-policy', includes_value='same-origin')
        self.metadata.http_unsafe_referrer = self.header_exists(name='referrer-policy', includes_value='unsafe-url')
        self.metadata.http_xss_protection = self.header_exists(name='x-xss-protection', includes_value='1; mode=block')
        http_statuses = [505]
        if transport.http1_support:
            http_statuses.append(transport.http1_code)
        if transport.http1_1_support:
            http_statuses.append(transport.http1_1_code)
        if transport.client_certificate_expected:
            http_statuses.append(511)
        self.metadata.http_status_code = min(http_statuses)
        self.metadata.http1_support = transport.http1_support
        self.metadata.http1_1_support = transport.http1_1_support
        self.metadata.http2_support = transport.http2_support
        self.http2_cleartext_support = transport.http2_cleartext_support

    def header_exists(self, name :str, includes_value :str = None) -> bool:
        if not isinstance(name, str):
            raise AttributeError(f'Invalid value for name, got {type(name)} expected str')
        checks = []
        if includes_value is not None:
            if not isinstance(includes_value, str):
                raise AttributeError(f'Invalid value for includes_value, got {type(includes_value)} expected str')
            checks.append(self.transport.http1_support and name in self.transport.http1_headers and includes_value in self.transport.http1_headers[name])
            checks.append(self.transport.http1_1_support and name in self.transport.http1_1_headers and includes_value in self.transport.http1_1_headers[name])
        else:
            checks.append(self.transport.http1_support and name in self.transport.http1_headers)
            checks.append(self.transport.http1_1_support and name in self.transport.http1_1_headers)
        return any(checks)

    def parse_openssl_errors(self, errors :list[tuple[X509, int]]):
        if not isinstance(errors, list):
            return
        for cert, errno in errors:
            message = exceptions.X509_MESSAGES[errno]
            if errno in exceptions.X509_MESSAGES and self.x509.get_serial_number() == cert.get_serial_number() and message not in self.certificate_verify_messages:
                self.certificate_verify_messages.append(exceptions.X509_MESSAGES[errno])

    def pcidss_compliant(self) -> bool:
        super().pcidss_compliant()
        self.validation_checks[pci.VALIDATION_WEAK_CIPHER] = not self.metadata.weak_cipher or self.metadata.negotiated_cipher_bits >= pci.WEAK_CIPHER_BITS
        if self.validation_checks[pci.VALIDATION_WEAK_CIPHER] is False:
            self.certificate_verify_messages.append(pci.PCIDSS_NON_COMPLIANCE_CIPHER)
        self.validation_checks[pci.VALIDATION_WEAK_PROTOCOL] = self.metadata.negotiated_protocol not in util.WEAK_PROTOCOL.keys()
        if self.validation_checks[pci.VALIDATION_WEAK_PROTOCOL] is False:
            self.certificate_verify_messages.append(pci.PCIDSS_NON_COMPLIANCE_WEAK_PROTOCOL)
        self.validation_checks[pci.VALIDATION_DEPRECATED_ALGO] = not self.metadata.dnssec_algorithm or self.metadata.dnssec_algorithm not in util.WEAK_DNSSEC_ALGORITHMS.keys()
        if self.validation_checks[pci.VALIDATION_DEPRECATED_ALGO] is False:
            self.certificate_verify_messages.append(pci.PCIDSS_NON_COMPLIANCE_WEAK_ALGORITHMS)
        self.validation_checks[pci.VALIDATION_KNOWN_VULN_SESSION_RESUMPTION] = all([
            not self.metadata.session_resumption_tickets or util.PROTOCOL_TEXT_MAP[self.metadata.negotiated_protocol] == SSL.TLS1_3_VERSION,
            not self.metadata.session_resumption_caching or util.PROTOCOL_TEXT_MAP[self.metadata.negotiated_protocol] == SSL.TLS1_3_VERSION,
        ])
        self.validation_checks[pci.VALIDATION_KNOWN_VULN_RENEGOTIATION] = not self.metadata.client_renegotiation
        self.validation_checks[pci.VALIDATION_KNOWN_VULN_COMPRESSION] = not self.metadata.compression_support
        if any([
                self.validation_checks[pci.VALIDATION_KNOWN_VULN_RENEGOTIATION],
                self.validation_checks[pci.VALIDATION_KNOWN_VULN_COMPRESSION],
                self.validation_checks[pci.VALIDATION_KNOWN_VULN_SESSION_RESUMPTION],
            ]):
            self.certificate_verify_messages.append(pci.PCIDSS_NON_COMPLIANCE_KNOWN_VULNERABILITIES)

    def fips_compliant(self) -> bool:
        super().fips_compliant()
        if util.PROTOCOL_TEXT_MAP[self.metadata.negotiated_protocol] not in [SSL.TLS1_2_VERSION, SSL.TLS1_3_VERSION]:
            self.validation_checks[fips.VALIDATION_MTLS] = not self.metadata.client_certificate_expected

        self.validation_checks[fips.VALIDATION_WEAK_PROTOCOL] = self.metadata.negotiated_protocol not in ['SSLv2 (0x02ff)', 'SSLv3 (0x0300)']
        if self.validation_checks[fips.VALIDATION_WEAK_PROTOCOL] is False:
            self.certificate_verify_messages.append(fips.FIPS_NON_COMPLIANCE_WEAK_PROTOCOL)
            self.validation_checks[fips.VALIDATION_WEAK_CIPHER] = False
            self.certificate_verify_messages.append(fips.FIPS_NON_COMPLIANCE_CIPHER)
        else:
            ok_cipher = self.metadata.negotiated_cipher_bits >= fips.WEAK_CIPHER_BITS
            if ok_cipher and util.PROTOCOL_TEXT_MAP[self.metadata.negotiated_protocol] == SSL.TLS1_2_VERSION:
                ok_cipher = all(c in self.metadata.negotiated_cipher for c in fips.ALLOWED_CIPHERS)
            if ok_cipher and util.PROTOCOL_TEXT_MAP[self.metadata.negotiated_protocol] in [SSL.TLS1_VERSION, SSL.TLS1_1_VERSION]:
                ok_cipher = all(c in self.metadata.negotiated_cipher for c in fips.ALLOWED_DEPRECATED_TLS_CIPHERS)
            self.validation_checks[fips.VALIDATION_WEAK_CIPHER] = ok_cipher
            if self.validation_checks[fips.VALIDATION_WEAK_CIPHER] is False:
                self.certificate_verify_messages.append(fips.FIPS_NON_COMPLIANCE_CIPHER)

    def nist_compliant(self) -> bool:
        # sourcery skip: extract-duplicate-method, extract-method, remove-redundant-if, split-or-ifs
        # sourcery will add bloat here, this is massively more concise logic then theirs
        super().nist_compliant()
        if util.PROTOCOL_TEXT_MAP[self.metadata.negotiated_protocol] in [SSL.TLS1_2_VERSION, SSL.TLS1_3_VERSION]:
            self.validation_checks[nist.VALIDATION_WEAK_PROTOCOL] = True
            self.validation_checks[nist.VALIDATION_MTLS] = self.metadata.client_certificate_expected
            ok_cipher = self.metadata.negotiated_cipher_bits >= nist.WEAK_CIPHER_BITS
            if ok_cipher and util.PROTOCOL_TEXT_MAP[self.metadata.negotiated_protocol] == SSL.TLS1_2_VERSION:
                ok_cipher = all(c in self.metadata.negotiated_cipher for c in nist.ALLOWED_CIPHERS)
            self.validation_checks[nist.VALIDATION_WEAK_CIPHER] = ok_cipher
            if self.validation_checks[nist.VALIDATION_WEAK_CIPHER] is False:
                self.certificate_verify_messages.append(nist.NIST_NON_COMPLIANCE_CIPHER)
        else:
            self.validation_checks[nist.VALIDATION_WEAK_PROTOCOL] = False
            self.certificate_verify_messages.append(nist.NIST_NON_COMPLIANCE_WEAK_PROTOCOL)
            self.validation_checks[nist.VALIDATION_WEAK_CIPHER] = False
            self.certificate_verify_messages.append(nist.NIST_NON_COMPLIANCE_CIPHER)

    def verify(self) -> bool:
        super().verify()
        tldext = TLDExtract(cache_dir='/tmp')(f'http://{self.metadata.host}')
        ca = util.get_basic_constraints(self.certificate)
        logger.debug('Server certificate validations')
        self.validation_checks[VALIDATION_BASIC_CONSTRAINTS_CA] = True
        if self.transport.client_certificate_expected:
            self.validation_checks[VALIDATION_CLIENT_AUTH_USAGE] = isinstance(self.transport.client_certificate, X509) and util.key_usage_exists(self.transport.client_certificate.to_cryptography(), 'clientAuth') is True
            self.validation_checks[VALIDATION_CLIENT_AUTHENTICATION] = self.transport.client_certificate_match and isinstance(self.transport.client_certificate, X509) and self.transport.negotiated_protocol is not None and util.key_usage_exists(self.certificate, 'clientAuth') is True
            if self.validation_checks[VALIDATION_CLIENT_AUTHENTICATION] is False:
                self.certificate_verify_messages.append(exceptions.VALIDATION_ERROR_CLIENT_AUTHENTICATION)
        if isinstance(ca, bool) and ca is True:
            self.validation_checks[VALIDATION_BASIC_CONSTRAINTS_CA] = False
            self.certificate_verify_messages.append('Server (leaf) certificates should not be a CA, it could enable impersonation attacks')
        self.validation_checks[VALIDATION_VALID_TLS_USAGE] = util.key_usage_exists(self.certificate, 'digital_signature') is True and util.key_usage_exists(self.certificate, 'serverAuth') is True
        self.validation_checks[VALIDATION_SUBJECT_CN_VALID] = util.validate_common_name(self.metadata.certificate_common_name, self.metadata.host) is True
        self.validation_checks[VALIDATION_MATCH_HOSTNAME] = util.match_hostname(self.metadata.host, self.certificate)
        self.metadata.certificate_is_self_signed = util.is_self_signed(self.certificate)
        self.validation_checks[VALIDATION_NOT_SELF_SIGNED] = self.metadata.certificate_is_self_signed is False
        if self.metadata.preferred_protocol != self.metadata.negotiated_protocol:
            self.certificate_verify_messages.append(exceptions.VALIDATION_ERROR_SCSV.format(protocol=self.metadata.preferred_protocol, fallback=self.metadata.negotiated_protocol))
        if self.metadata.certificate_is_self_signed:
            self.validation_checks[VALIDATION_ROOT_CA_TRUST] = False
            self.certificate_verify_messages.append('The CA is not properly imported as a trusted CA into the browser, Chrome based browsers will block visitors and show them ERR_CERT_AUTHORITY_INVALID')
        self.validation_checks[VALIDATION_DEPRECATED_TLS_PROTOCOLS] = self.metadata.negotiated_protocol not in util.WEAK_PROTOCOL.keys()
        if self.validation_checks[VALIDATION_DEPRECATED_TLS_PROTOCOLS] is False:
            self.certificate_verify_messages.append(util.WEAK_PROTOCOL[self.metadata.negotiated_protocol])
        if self.metadata.revocation_ocsp_stapling is True:
            self.validation_checks[VALIDATION_OCSP_STAPLE_SATISFIED] = False
            if self.metadata.revocation_ocsp_response is not None:
                self.validation_checks[VALIDATION_OCSP_STAPLE_SATISFIED] = True
        if self.metadata.revocation_ocsp_must_staple is True:
            self.validation_checks[VALIDATION_OCSP_MUST_STAPLE_SATISFIED] = False
            if self.metadata.revocation_ocsp_status == util.OCSP_CERT_STATUS[0]:
                self.validation_checks[VALIDATION_OCSP_MUST_STAPLE_SATISFIED] = True
        if self.metadata.certification_authority_authorization:
            self.validation_checks[VALIDATION_VALID_CAA] = util.caa_valid(self.metadata.host, self.x509, self.certificate_chain)
        if self.metadata.dnssec:
            self.validation_checks[VALIDATION_VALID_DNSSEC] = util.dnssec_valid(self.metadata.host)
            if self.validation_checks[VALIDATION_VALID_DNSSEC] is False and tldext.registered_domain != self.metadata.host:
                self.validation_checks[VALIDATION_VALID_DNSSEC] = util.dnssec_valid(tldext.registered_domain)
                if self.validation_checks[VALIDATION_VALID_DNSSEC]:
                    self.certificate_verify_messages.append(exceptions.VALIDATION_ERROR_DNSSEC_TARGET)
                else:
                    self.certificate_verify_messages.append(exceptions.VALIDATION_ERROR_DNSSEC_REGISTERED_DOMAIN)
        else:
            self.certificate_verify_messages.append(exceptions.VALIDATION_ERROR_DNSSEC_MISSING)
        if self.metadata.certificate_validation_type == util.VALIDATION_TYPES['DV']:
            self.certificate_verify_messages.append(exceptions.VALIDATION_ERROR_CERTIFICATE_VALIDATION_TYPE)
        if not self.metadata.certification_authority_authorization:
            self.certificate_verify_messages.append(exceptions.VALIDATION_ERROR_CERTIFICATION_AUTHORITY_AUTHORIZATION)
        if self.metadata.session_resumption_caching and util.PROTOCOL_TEXT_MAP[self.metadata.negotiated_protocol] != SSL.TLS1_3_VERSION:
            self.certificate_verify_messages.append(exceptions.VALIDATION_ERROR_SESSION_RESUMPTION_CACHING)
        if self.metadata.session_resumption_caching and util.PROTOCOL_TEXT_MAP[self.metadata.negotiated_protocol] == SSL.TLS1_3_VERSION:
            self.certificate_verify_messages.append(exceptions.VALIDATION_ERROR_SESSION_RESUMPTION_CACHING_TLS1_3)
        if self.metadata.session_resumption_tickets and util.PROTOCOL_TEXT_MAP[self.metadata.negotiated_protocol] != SSL.TLS1_3_VERSION:
            self.certificate_verify_messages.append(exceptions.VALIDATION_ERROR_SESSION_RESUMPTION_TICKETS)
        if self.metadata.client_renegotiation:
            self.certificate_verify_messages.append(exceptions.VALIDATION_ERROR_CLIENT_RENEGOTIATION)
        if self.metadata.compression_support:
            self.certificate_verify_messages.append(exceptions.VALIDATION_ERROR_COMPRESSION_SUPPORT)
        self.validation_checks[VALIDATION_DEPRECATED_DNSSEC_ALGO] = self.metadata.dnssec_algorithm not in util.WEAK_DNSSEC_ALGORITHMS.keys()
        if self.validation_checks[VALIDATION_DEPRECATED_DNSSEC_ALGO] is False:
            self.certificate_verify_messages.append(util.WEAK_DNSSEC_ALGORITHMS[self.metadata.dnssec_algorithm])
        interference_versions = ''.join(self.metadata.tls_version_interference_versions)
        current_version = 'TLSv1.3'
        if '0x304' in interference_versions:
            self.certificate_verify_messages.append(exceptions.VALIDATION_ERROR_VERSION_INTERFERENCE_CURRENT.format(current_version=current_version))
        if '0x303' in interference_versions:
            common_version = 'TLSv1.2'
            self.certificate_verify_messages.append(exceptions.VALIDATION_ERROR_VERSION_INTERFERENCE_COMMON.format(current_version=current_version, common_version=common_version))
        if '0x302' in interference_versions:
            old_version = 'TLSv1.1'
            self.certificate_verify_messages.append(exceptions.VALIDATION_ERROR_VERSION_INTERFERENCE_OLD.format(old_version=old_version))
        elif any(['0x301' in interference_versions, '0x300' in interference_versions, '0x2ff' in interference_versions]):
            self.certificate_verify_messages.append(exceptions.VALIDATION_ERROR_VERSION_INTERFERENCE_OBSOLETE)

        return self.certificate_valid

    def _get_root_certs(self, trust_store :TrustStore):
        contexts = [
            context.PLATFORM_ANDROID7,
            context.PLATFORM_ANDROID8,
            context.PLATFORM_ANDROID9,
            context.PLATFORM_ANDROID10,
            context.PLATFORM_ANDROID11,
            context.PLATFORM_ANDROID12,
            context.SOURCE_CCADB,
            context.SOURCE_JAVA,
            context.SOURCE_LINUX,
            context.SOURCE_CERTIFI
        ]
        for context_type in contexts:
            try:
                yield trust_store.get_certificate_from_store(context_type=context_type)
            except FileExistsError:
                pass

    def _validate_roots(self, trust_store :TrustStore):
        for cert in self._get_root_certs(trust_store):
            if cert.get_serial_number() in self._root_certs:
                continue
            root_validator = RootCertValidator()
            root_validator.init_x509(cert)
            root_validator.metadata.certificate_root_ca = True
            root_validator.verify()
            root_validator.verify_trust(trust_store)
            root_validator.pcidss_compliant()
            root_validator.fips_compliant()
            root_validator.nist_compliant()
            self.peer_validations.append(root_validator)
            self._root_certs.append(cert.get_serial_number())

    def verify_chain(self, progress_bar :callable = lambda *args: None) -> bool:
        if not self._pem_certificate_chain:
            return False

        validator_key_usage, validator_extended_key_usage = util.gather_key_usages(self.certificate)
        validation_result = f'Validated: {",".join(validator_key_usage + validator_extended_key_usage)}'
        try:
            logger.debug('certificate chain validation')
            util.validate_certificate_chain(self._der, self._pem_certificate_chain, validator_key_usage, validator_extended_key_usage)
        except RevokedError as ex:
            logger.debug(ex, stack_info=True)
            self.validation_checks[VALIDATION_REVOCATION] = False
            self.certificate_chain_valid = False
            validation_result = str(ex)
        except InvalidCertificateError as ex:
            logger.debug(ex, stack_info=True)
            self.certificate_chain_valid = False
            validation_result = str(ex)
        except PathValidationError as ex:
            logger.debug(ex, stack_info=True)
            self.certificate_chain_valid = False
            validation_result = str(ex)
        except PathBuildingError as ex:
            logger.debug(ex, stack_info=True)
            self.certificate_chain_valid = False
            validation_result = str(ex)
        except Exception as ex:
            logger.warning(ex, stack_info=True)
            validation_result = str(ex)

        self.certificate_chain_validation_result = validation_result
        self.certificate_chain_valid = self.certificate_chain_valid is not False

        progress_bar()
        def map_peers(peers :list[X509]):
            peers_map = {}
            for cert in peers:
                ski, aki = util.get_ski_aki(cert.to_cryptography())
                if ski is None or aki is None:
                    logger.warning(f'Certificate {cert.get_subject()} has no SKI or AKI')
                peers_map[ski] = [cert, aki]
            return peers_map

        peer_lookup = map_peers(self.certificate_chain)
        for cert in self.certificate_chain:
            progress_bar()
            if cert.get_serial_number() == self.x509.get_serial_number():
                continue
            ca, _ = util.get_basic_constraints(cert.to_cryptography())
            peer_validator = PeerCertValidator()
            peer_validator.init_x509(cert)
            peer_validator.metadata.certificate_intermediate_ca = ca is True
            trust_store = None
            checked_peer_as_root = False
            if peer_validator.metadata.certificate_authority_key_identifier is None:
                trust_store = TrustStore(authority_key_identifier=peer_validator.metadata.certificate_subject_key_identifier)
                checked_peer_as_root = True
            elif peer_validator.metadata.certificate_authority_key_identifier not in peer_lookup.keys():
                trust_store = TrustStore(authority_key_identifier=peer_validator.metadata.certificate_authority_key_identifier)
            if isinstance(trust_store, TrustStore):
                logger.info(f'Checking for Root CA issuer {cert.get_issuer()} with SKI {peer_validator.metadata.certificate_authority_key_identifier}')
                self._root_certs = []
                self.validation_checks[VALIDATION_ROOT_CA_TRUST] = trust_store.is_trusted
                if any([
                        trust_store.exists(context.SOURCE_CCADB),
                        trust_store.exists(context.SOURCE_JAVA),
                        trust_store.exists(context.SOURCE_LINUX),
                        trust_store.exists(context.SOURCE_CERTIFI),
                        trust_store.exists(context.PLATFORM_ANDROID7),
                        trust_store.exists(context.PLATFORM_ANDROID8),
                        trust_store.exists(context.PLATFORM_ANDROID9),
                        trust_store.exists(context.PLATFORM_ANDROID10),
                        trust_store.exists(context.PLATFORM_ANDROID11),
                        trust_store.exists(context.PLATFORM_ANDROID12),
                    ]):
                    self._validate_roots(trust_store)
                if not self.validation_checks[VALIDATION_ROOT_CA_TRUST]:
                    self.certificate_verify_messages.append(exceptions.VALIDATION_ERROR_MISSING_ROOT_CA_AKI.format(serial_number=peer_validator.metadata.certificate_serial_number_hex))
                else:
                    checked_peer_as_root = checked_peer_as_root is True

            if not checked_peer_as_root:
                peer_validator.verify()
                self.validation_checks[VALIDATION_REVOCATION] = self.validation_checks.get(VALIDATION_REVOCATION) is not False
                peer_validator.pcidss_compliant()
                peer_validator.fips_compliant()
                peer_validator.nist_compliant()
                self.peer_validations.append(peer_validator)

        return all([self.certificate_valid, self.certificate_chain_valid])

    def extract_x509_metadata(self, x509 :X509):
        super().extract_x509_metadata(x509)
        self.metadata.certification_authority_authorization = util.caa_exist(self.metadata.host)
        answer :list[RRset] = util.get_dnssec_answer(self.metadata.host)
        if answer:
            self.metadata.dnssec = True
            _, _, _, _, _, _, algorithm, *rest = answer[0].to_text().split()
            self.metadata.dnssec_algorithm = util.DNSSEC_ALGORITHMS[int(algorithm)]
